import dateutil.tz
import pytest
from datetime import datetime, timedelta
from collections import OrderedDict

from indy_node.test.auth_rule.auth_framework.disable_taa import TAADisableTest
from indy_node.test.auth_rule.auth_framework.edit_frozen_ledgers import EditFrozenLedgersTest
from plenum.common.constants import STEWARD, TRUSTEE, IDENTITY_OWNER

from indy_common.constants import (
    ENDORSER, START, NETWORK_MONITOR, NETWORK_MONITOR_STRING
)
from indy_common.authorize import auth_map

from plenum.test.helper import randomText, view_change_timeout
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.testing_utils import FakeSomething

from indy_node.test.auth_rule.auth_framework.claim_def import AddClaimDefTest, EditClaimDefTest
from indy_node.test.auth_rule.auth_framework.add_roles import AddNewTrusteeTest, AddNewStewardTest, \
    AddNewEndorserTest, AddNewNetworkMonitorTest, AddNewIdentityOwnerTest
from indy_node.test.auth_rule.auth_framework.edit_roles import EditTrusteeToStewardTest, \
    EditTrusteeToEndorserTest, EditTrusteeToNetworkMonitorTest, EditTrusteeToIdentityOwnerTest, \
    EditStewardToTrusteeTest, EditStewardToEndorserTest, EditStewardToNetworkMonitorTest, \
    EditStewardToIdentityOwnerTest, EditEndorserToTrusteeTest, EditEndorserToStewardTest, \
    EditEndorserToIdentityOwnerTest, EditEndorserToNetworkMonitorTest, EditIdentityOwnerToNetworkMonitorTest, \
    EditIdentityOwnerToTrusteeTest, EditIdentityOwnerToEndorserTest, EditIdentityOwnerToStewardTest, \
    EditNetworkMonitorToIdentityOwnerTest, EditNetworkMonitorToTrusteeTest, EditNetworkMonitorToStewardTest, \
    EditNetworkMonitorToEndorserTest, EditStewardToStewardTest, EditTrusteeToTrusteeTest, \
    EditEndorserToEndorserTest, EditNetworkMonitorToNetworkMonitorTest, EditIdentityOwnerToIdentityOwnerTest
from indy_node.test.auth_rule.auth_framework.key_rotation import RotateKeyTest
from indy_node.test.auth_rule.auth_framework.schema import SchemaTest
from indy_node.test.auth_rule.auth_framework.upgrade import StartUpgradeTest, CancelUpgradeTest
from indy_node.test.upgrade.helper import bumpedVersion
from indy_node.test.auth_rule.auth_framework.add_attrib import AddAttribTest
from indy_node.test.auth_rule.auth_framework.edit_attrib import EditAttribTest
from indy_node.test.auth_rule.auth_framework.edit_auth_rules import AuthRulesTest
from indy_node.test.auth_rule.auth_framework.edit_auth_rule import AuthRuleTest
from indy_node.test.auth_rule.auth_framework.node_services import AddNewNodeTest, AddNewNodeEmptyServiceTest, \
    DemoteNodeTest, PromoteNodeTest
from indy_node.test.auth_rule.auth_framework.node_properties import EditNodeIpTest, EditNodePortTest, \
    EditNodeClientIpTest, EditNodeClientPortTest, EditNodeBlsTest
from indy_node.test.auth_rule.auth_framework.pool_config import PoolConfigTest
from indy_node.test.auth_rule.auth_framework.restart import RestartTest
from indy_node.test.auth_rule.auth_framework.revoc_reg_def import AddRevocRegDefTest, \
    EditRevocRegDefTest
from indy_node.test.auth_rule.auth_framework.revoc_reg_entry import AddRevocRegEntryTest, EditRevocRegEntryTest
from indy_node.test.auth_rule.auth_framework.txn_author_agreement import TxnAuthorAgreementTest
from indy_node.test.auth_rule.auth_framework.txn_author_agreement_aml import TxnAuthorAgreementAMLTest
from indy_node.test.auth_rule.auth_framework.validator_info import ValidatorInfoTest
from indy_node.test.pool_config.conftest import poolConfigWTFF
from indy_node.test.upgrade.conftest import patch_packet_mgr_output, EXT_PKT_NAME, EXT_PKT_VERSION

nodeCount = 7

from stp_core.common.log import Logger

Logger().enableStdLogging()


@pytest.fixture(scope="module")
def tconf(tconf):
    with view_change_timeout(tconf, vc_timeout=5) as tconf:
        yield tconf


class TestAuthRuleUsing():
    map_of_tests = OrderedDict({
        auth_map.add_new_trustee.get_action_id(): AddNewTrusteeTest,
        auth_map.add_new_steward.get_action_id(): AddNewStewardTest,
        auth_map.add_new_endorser.get_action_id(): AddNewEndorserTest,
        auth_map.add_new_network_monitor.get_action_id(): AddNewNetworkMonitorTest,
        auth_map.add_new_identity_owner.get_action_id(): AddNewIdentityOwnerTest,
        auth_map.add_attrib.get_action_id(): AddAttribTest,
        auth_map.edit_attrib.get_action_id(): EditAttribTest,
        auth_map.add_revoc_reg_def.get_action_id(): AddRevocRegDefTest,
        auth_map.edit_revoc_reg_def.get_action_id(): EditRevocRegDefTest,
        auth_map.add_revoc_reg_entry.get_action_id(): AddRevocRegEntryTest,
        auth_map.edit_revoc_reg_entry.get_action_id(): EditRevocRegEntryTest,
        auth_map.edit_role_actions[TRUSTEE][STEWARD].get_action_id(): EditTrusteeToStewardTest,
        auth_map.edit_role_actions[TRUSTEE][ENDORSER].get_action_id(): EditTrusteeToEndorserTest,
        auth_map.edit_role_actions[TRUSTEE][NETWORK_MONITOR].get_action_id(): EditTrusteeToNetworkMonitorTest,
        auth_map.edit_role_actions[TRUSTEE][IDENTITY_OWNER].get_action_id(): EditTrusteeToIdentityOwnerTest,
        auth_map.edit_role_actions[STEWARD][TRUSTEE].get_action_id(): EditStewardToTrusteeTest,
        auth_map.edit_role_actions[STEWARD][ENDORSER].get_action_id(): EditStewardToEndorserTest,
        auth_map.edit_role_actions[STEWARD][NETWORK_MONITOR].get_action_id(): EditStewardToNetworkMonitorTest,
        auth_map.edit_role_actions[STEWARD][IDENTITY_OWNER].get_action_id(): EditStewardToIdentityOwnerTest,
        auth_map.edit_role_actions[ENDORSER][TRUSTEE].get_action_id(): EditEndorserToTrusteeTest,
        auth_map.edit_role_actions[ENDORSER][STEWARD].get_action_id(): EditEndorserToStewardTest,
        auth_map.edit_role_actions[ENDORSER][NETWORK_MONITOR].get_action_id(): EditEndorserToNetworkMonitorTest,
        auth_map.edit_role_actions[ENDORSER][IDENTITY_OWNER].get_action_id(): EditEndorserToIdentityOwnerTest,
        auth_map.edit_role_actions[IDENTITY_OWNER][TRUSTEE].get_action_id(): EditIdentityOwnerToTrusteeTest,
        auth_map.edit_role_actions[IDENTITY_OWNER][STEWARD].get_action_id(): EditIdentityOwnerToStewardTest,
        auth_map.edit_role_actions[IDENTITY_OWNER][ENDORSER].get_action_id(): EditIdentityOwnerToEndorserTest,
        auth_map.edit_role_actions[IDENTITY_OWNER][NETWORK_MONITOR].get_action_id(): EditIdentityOwnerToNetworkMonitorTest,
        auth_map.edit_role_actions[NETWORK_MONITOR][TRUSTEE].get_action_id(): EditNetworkMonitorToTrusteeTest,
        auth_map.edit_role_actions[NETWORK_MONITOR][STEWARD].get_action_id(): EditNetworkMonitorToStewardTest,
        auth_map.edit_role_actions[NETWORK_MONITOR][ENDORSER].get_action_id(): EditNetworkMonitorToEndorserTest,
        auth_map.edit_role_actions[NETWORK_MONITOR][IDENTITY_OWNER].get_action_id(): EditNetworkMonitorToIdentityOwnerTest,
        auth_map.edit_role_actions[TRUSTEE][TRUSTEE].get_action_id(): EditTrusteeToTrusteeTest,
        auth_map.edit_role_actions[STEWARD][STEWARD].get_action_id(): EditStewardToStewardTest,
        auth_map.edit_role_actions[ENDORSER][ENDORSER].get_action_id(): EditEndorserToEndorserTest,
        auth_map.edit_role_actions[NETWORK_MONITOR][NETWORK_MONITOR].get_action_id(): EditNetworkMonitorToNetworkMonitorTest,
        auth_map.edit_role_actions[IDENTITY_OWNER][IDENTITY_OWNER].get_action_id(): EditIdentityOwnerToIdentityOwnerTest,
        auth_map.key_rotation.get_action_id(): RotateKeyTest,
        auth_map.txn_author_agreement_aml.get_action_id(): TxnAuthorAgreementAMLTest,
        auth_map.add_schema.get_action_id(): SchemaTest,
        auth_map.add_claim_def.get_action_id(): AddClaimDefTest,
        auth_map.edit_claim_def.get_action_id(): EditClaimDefTest,
        auth_map.start_upgrade.get_action_id(): StartUpgradeTest,
        auth_map.cancel_upgrade.get_action_id(): CancelUpgradeTest,
        auth_map.pool_restart.get_action_id(): RestartTest,
        auth_map.pool_config.get_action_id(): PoolConfigTest,
        auth_map.auth_rule.get_action_id(): AuthRuleTest,
        auth_map.auth_rules.get_action_id(): AuthRulesTest,
        auth_map.validator_info.get_action_id(): ValidatorInfoTest,
        auth_map.adding_new_node.get_action_id(): AddNewNodeTest,
        auth_map.adding_new_node_with_empty_services.get_action_id(): AddNewNodeEmptyServiceTest,
        auth_map.demote_node.get_action_id(): DemoteNodeTest,
        auth_map.promote_node.get_action_id(): PromoteNodeTest,
        auth_map.change_node_ip.get_action_id(): EditNodeIpTest,
        auth_map.change_node_port.get_action_id(): EditNodePortTest,
        auth_map.change_client_ip.get_action_id(): EditNodeClientIpTest,
        auth_map.change_client_port.get_action_id(): EditNodeClientPortTest,
        auth_map.change_bls_key.get_action_id(): EditNodeBlsTest,
        auth_map.disable_txn_author_agreement.get_action_id(): TAADisableTest,
        auth_map.edit_frozen_ledgers.get_action_id(): EditFrozenLedgersTest,
    })

    # TODO a workaround until sdk aceepts empty TAA to make possible its deactivation
    map_of_tests[auth_map.txn_author_agreement.get_action_id()] = TxnAuthorAgreementTest

    @pytest.fixture(scope='module')
    def pckg(self):
        return (EXT_PKT_NAME, EXT_PKT_VERSION)

    @pytest.fixture(scope='module')
    def monkeymodule(self):
        from _pytest.monkeypatch import MonkeyPatch
        mpatch = MonkeyPatch()
        yield mpatch
        mpatch.undo()

    @pytest.fixture(scope='module')
    def validUpgrade(self, nodeIds, tconf, pckg, monkeymodule):
        schedule = {}
        unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
        startAt = unow + timedelta(seconds=3000)
        acceptableDiff = tconf.MinSepBetweenNodeUpgrades + 1
        for i in nodeIds:
            schedule[i] = datetime.isoformat(startAt)
            startAt = startAt + timedelta(seconds=acceptableDiff + 3)

        new_version = bumpedVersion(pckg[1])
        patch_packet_mgr_output(monkeymodule, pckg[0], pckg[1], new_version)

        return dict(name='upgrade-{}'.format(randomText(3)), version=new_version,
                    action=START, schedule=schedule, timeout=1, package=pckg[0],
                    sha256='db34a72a90d026dae49c3b3f0436c8d3963476c77468ad955845a1ccf7b03f55')

    @pytest.fixture(scope="module")
    def sdk_wallet_network_monitor(self, looper, sdk_pool_handle, sdk_wallet_trustee):
        return sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                               alias='NM-1', role=NETWORK_MONITOR_STRING)

    @pytest.fixture(scope="module")  # noqa: F811
    def env(self,
            looper,
            tconf,
            tdir,
            sdk_pool_handle,
            sdk_wallet_trustee,
            sdk_wallet_steward,
            sdk_wallet_endorser,
            sdk_wallet_client,
            validUpgrade,
            poolConfigWTFF,
            sdk_wallet_network_monitor,
            txnPoolNodeSet):
        role_to_wallet = {
            TRUSTEE: sdk_wallet_trustee,
            STEWARD: sdk_wallet_steward,
            ENDORSER: sdk_wallet_endorser,
            NETWORK_MONITOR: sdk_wallet_network_monitor,
            IDENTITY_OWNER: sdk_wallet_client,
        }
        return FakeSomething(looper=looper,
                             tconf=tconf,
                             tdir=tdir,
                             sdk_pool_handle=sdk_pool_handle,
                             sdk_wallet_trustee=sdk_wallet_trustee,
                             sdk_wallet_steward=sdk_wallet_steward,
                             sdk_wallet_client=sdk_wallet_client,
                             role_to_wallet=role_to_wallet,
                             txnPoolNodeSet=txnPoolNodeSet,
                             valid_upgrade=validUpgrade,
                             pool_config_wtff=poolConfigWTFF)

    @pytest.fixture(scope='module', params=[k for k in map_of_tests.keys()])
    def auth_rule_tests(self, request, env):
        action_id = request.param
        test_cls = self.map_of_tests[action_id]
        test = test_cls(env, action_id)
        return action_id, test

    def test_auth_rule_using(self, auth_rule_tests):
        descr, test = auth_rule_tests
        print("Running test: {}".format(descr))
        test.prepare()
        test.run()
        test.result()
        test.down()



from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.constants import ACTION, POOL_RESTART


def test_pool_restart_add_action(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=POOL_RESTART,
                                                                 field=ACTION,
                                                                 value='start',
                                                                 is_owner=is_owner)])



from indy_node.test import waits
from indy_node.test.upgrade.helper import checkUpgradeScheduled, \
    sdk_send_upgrade
from plenum.common.constants import VERSION
from plenum.test.delayers import req_delay, ppDelay, pDelay, cDelay, ppgDelay
from plenum.test.test_node import getNonPrimaryReplicas
from stp_core.loop.eventually import eventually


def test_node_handles_forced_upgrade_on_propagate(
        looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee,
        validUpgradeExpForceTrue):
    """
    Verifies that POOL_UPGRADE force=true request is handled immediately when
    the node receives it in a PROPAGATE from any other node
    """
    slow_node = getNonPrimaryReplicas(nodeSet, instId=0)[-1].node

    # Stash all except PROPAGATEs from Gamma
    slow_node.clientIbStasher.delay(req_delay())
    slow_node.nodeIbStasher.delay(ppgDelay(sender_filter='Alpha'))
    slow_node.nodeIbStasher.delay(ppgDelay(sender_filter='Beta'))
    slow_node.nodeIbStasher.delay(ppDelay())
    slow_node.nodeIbStasher.delay(pDelay())
    slow_node.nodeIbStasher.delay(cDelay())

    sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee,
                     validUpgradeExpForceTrue)

    looper.run(eventually(checkUpgradeScheduled,
                          [slow_node],
                          validUpgradeExpForceTrue[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))



import os
import shutil

from indy_common.setup_util import Setup

BASE_DIR = os.path.join(os.path.expanduser("~"), ".indy")
Setup(BASE_DIR).setupCommon()
Setup(BASE_DIR).setupNode()
Setup(BASE_DIR).setupClient()
if not os.path.exists(os.path.join(BASE_DIR, 'nssm.exe')):
    shutil.copy2(os.path.join(BASE_DIR, 'nssm_original.exe'),
                 os.path.join(BASE_DIR, 'nssm.exe'))



from plenum.common.util import get_utc_epoch
from storage.kv_in_memory import KeyValueStorageInMemory
from indy_node.persistence.idr_cache import IdrCache

identifier = "fake_identifier"
committed_items = (0,   # seq_no
                   get_utc_epoch(), # txn_time
                   "committed_ta_value",
                   "committed_role_value",
                   "committed_verkey_value",)
uncommitted_items = (1,
                     get_utc_epoch(), # txn_time
                     "uncommitted_ta_value",
                     "uncommitted_role_value",
                     "uncommitted_verkey_value",)


def make_idr_cache():
    kvs = KeyValueStorageInMemory()
    cache = IdrCache("TestCache", kvs)
    return cache


def test_committed():
    """
    Check that it is possible to set and get committed items
    """
    cache = make_idr_cache()
    cache.set(identifier, *committed_items)
    real_items = cache.get(identifier)
    assert committed_items == real_items
    real_items = cache.get(identifier, isCommitted=False)
    assert committed_items == real_items


def test_uncommitted():
    """
    Check that it is possible to set and get uncommitted items
    """
    cache = make_idr_cache()
    cache.set(identifier, *uncommitted_items, isCommitted=False)
    real_items = cache.get(identifier, isCommitted=False)
    assert uncommitted_items == real_items


def test_committed_and_uncommitted():
    """
    Check that uncommitted and committed can present together
    """
    cache = make_idr_cache()
    cache.set(identifier, *committed_items, isCommitted=True)
    cache.set(identifier, *uncommitted_items, isCommitted=False)
    assert uncommitted_items == cache.get(identifier, isCommitted=False)
    assert committed_items == cache.get(identifier, isCommitted=True)



import pytest
import json

from enum import Enum, unique

from indy.did import create_and_store_my_did, key_for_local_did

from plenum.common.constants import (
    TRUSTEE, STEWARD, NYM, TXN_TYPE, TARGET_NYM, VERKEY, ROLE,
    CURRENT_PROTOCOL_VERSION)
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.signer_did import DidSigner
from plenum.common.member.member import Member
from plenum.test.helper import sdk_gen_request, sdk_sign_request_objects

from indy_common.types import Request
from indy_common.roles import Roles

from indy_node.test.helper import createUuidIdentifierAndFullVerkey


#   TODO
#   - more specific string patterns for auth exc check


class DID(object):
    def __init__(self, did=None, role=Roles.IDENTITY_OWNER, verkey=None, creator=None, wallet_handle=None):
        self.did = did
        self.role = role
        self.verkey = verkey
        self.creator = creator
        self.wallet_handle = wallet_handle

    @property
    def wallet_did(self):
        return (self.wallet_handle, self.did)


@unique
class EnumBase(Enum):
    def __str__(self):
        return self.name


ActionIds = Enum('ActionIds', 'add edit', type=EnumBase)

# params for addition:
# - signer:
#   - role: Roles
# - dest:
#   - verkey in NYM: omitted, None, val
#   - role: Roles, omitted
NYMAddDestRoles = Enum(
    'NYMAddDestRoles',
    [(r.name, r.value) for r in Roles] + [('omitted', 'omitted')],
    type=EnumBase)

NYMAddDestVerkeys = Enum('NYMAddDestVerkeys', 'none val omitted', type=EnumBase)

# params for edition:
# - signer:
#   - [self, creator] + [other], where other = any of Roles
# - dest:
#   - role in ledger: Roles
#   - verkey in ledger: None, val
#   - role: Roles, omitted
#   - verkey in NYM: same, new (not None), demote(None), omitted
LedgerDIDVerkeys = Enum('LedgerDIDVerkeys', 'none val', type=EnumBase)
LedgerDIDRoles = Roles

NYMEditSignerTypes = Enum(
    'NYMEditSignerTypes',
    [(r.name, r.value) for r in Roles] + [('self', 'self'), ('creator', 'creator')],
    type=EnumBase
)

NYMEditDestRoles = NYMAddDestRoles
NYMEditDestVerkeys = Enum('NYMEditDestVerkeys', 'same new demote omitted', type=EnumBase)

dids = {}
did_editor_others = {}
did_provisioners = did_editor_others


# FIXTURES


@pytest.fixture(scope="module")
def trustee(looper, sdk_wallet_trustee):
    wh, did = sdk_wallet_trustee
    verkey = looper.loop.run_until_complete(key_for_local_did(wh, did))
    return DID(did=did, role=Roles.TRUSTEE, verkey=verkey, wallet_handle=wh)


@pytest.fixture(scope="module")
def poolTxnData(looper, poolTxnData, trustee):
    global dids
    global did_editor_others

    # TODO non-Trustee creators

    data = poolTxnData

    def _add_did(role, did_name, with_verkey=True):
        nonlocal data

        data['seeds'][did_name] = did_name + '0' * (32 - len(did_name))
        t_sgnr = DidSigner(seed=data['seeds'][did_name].encode())
        verkey = t_sgnr.full_verkey if with_verkey else None
        data['txns'].append(
            Member.nym_txn(nym=t_sgnr.identifier,
                           verkey=verkey,
                           role=role.value,
                           name=did_name,
                           creator=trustee.did)
        )

        if verkey:
            (sdk_did, sdk_verkey) = looper.loop.run_until_complete(
                create_and_store_my_did(
                    trustee.wallet_handle,
                    json.dumps({'seed': data['seeds'][did_name]}))
            )

        return DID(
            did=t_sgnr.identifier, role=role, verkey=verkey,
            creator=trustee, wallet_handle=trustee.wallet_handle
        )

    params = [(dr, dv) for dr in LedgerDIDRoles for dv in LedgerDIDVerkeys]
    for (dr, dv) in params:
        dids[(dr, dv)] = _add_did(
            dr, "{}-{}".format(dr.name, dv.name),
            with_verkey=(dv == LedgerDIDVerkeys.val)
        )

    for dr in Roles:
        did_editor_others[dr] = _add_did(dr, "{}-other".format(dr.name))

    return data


@pytest.fixture(scope="module", params=list(Roles))
def provisioner_role(request):
    return request.param


@pytest.fixture(scope="module")
def provisioner(poolTxnData, provisioner_role):
    return did_provisioners[provisioner_role]


@pytest.fixture(scope="module", params=list(NYMAddDestRoles))
def nym_add_dest_role(request):
    return request.param


@pytest.fixture(scope="module", params=list(NYMAddDestVerkeys))
def nym_add_dest_verkey(request):
    return request.param


@pytest.fixture(scope="function")
def add_op(nym_add_dest_role, nym_add_dest_verkey):
    did, verkey = createUuidIdentifierAndFullVerkey()

    op = {
        TXN_TYPE: NYM,
        TARGET_NYM: did,
        ROLE: nym_add_dest_role.value,
        VERKEY: verkey
    }

    if nym_add_dest_role == NYMAddDestRoles.omitted:
        del op[ROLE]

    if nym_add_dest_verkey == NYMAddDestVerkeys.omitted:
        del op[VERKEY]

    return op


@pytest.fixture(scope="module", params=list(LedgerDIDRoles))
def edited_ledger_role(request):
    return request.param


@pytest.fixture(scope="module", params=list(LedgerDIDVerkeys))
def edited_ledger_verkey(request):
    return request.param


@pytest.fixture(scope="function")
def edited(edited_ledger_role, edited_ledger_verkey):
    return dids[(edited_ledger_role, edited_ledger_verkey)]


@pytest.fixture(scope="module", params=list(NYMEditDestRoles))
def edited_nym_role(request):
    return request.param


@pytest.fixture(scope="module", params=list(NYMEditDestVerkeys))
def edited_nym_verkey(request):
    return request.param


@pytest.fixture(scope="module", params=list(NYMEditSignerTypes))
def editor_type(request):
    return request.param


@pytest.fixture(scope="function")
def editor(editor_type, edited):
    if editor_type == NYMEditSignerTypes.self:
        return edited
    elif editor_type == NYMEditSignerTypes.creator:
        return edited.creator
    else:
        return did_editor_others[Roles(editor_type.value)]


@pytest.fixture(scope="function")
def edit_op(edited, edited_nym_role, edited_nym_verkey):
    op = {
        TXN_TYPE: NYM,
        TARGET_NYM: edited.did,
    }

    if edited_nym_role != NYMEditDestRoles.omitted:
        op[ROLE] = edited_nym_role.value

    if edited_nym_verkey == NYMEditDestVerkeys.same:
        op[VERKEY] = edited.verkey
    elif edited_nym_verkey == NYMEditDestVerkeys.new:
        _, op[VERKEY] = createUuidIdentifierAndFullVerkey()
    elif edited_nym_verkey == NYMEditDestVerkeys.demote:
        if edited.verkey is None:
            return None  # pass that case since it is covered by `same` case as well
        else:
            op[VERKEY] = None

    return op


# TEST HELPERS

def auth_check(action_id, signer, op, did_ledger=None):
    op_role = Roles(op[ROLE]) if ROLE in op else None

    def check_promotion():
        # omitted role means IDENTITY_OWNER
        if op_role in (None, Roles.IDENTITY_OWNER):
            return signer.role in (Roles.TRUSTEE, Roles.STEWARD, Roles.ENDORSER)
        elif op_role in (Roles.TRUSTEE, Roles.STEWARD):
            return signer.role == Roles.TRUSTEE
        elif op_role in (Roles.ENDORSER, Roles.NETWORK_MONITOR):
            return signer.role in (Roles.TRUSTEE, Roles.STEWARD)

    def check_demotion():
        if did_ledger.role in (Roles.TRUSTEE, Roles.STEWARD):
            return signer.role == Roles.TRUSTEE
        elif did_ledger.role == Roles.ENDORSER:
            return (signer.role == Roles.TRUSTEE)
            # FIXME INDY-1968: uncomment when the task is addressed
            # return ((signer.role == Roles.TRUSTEE) or
            #        (signer.role == Roles.ENDORSER and
            #            is_self and is_owner))
        elif did_ledger.role == Roles.NETWORK_MONITOR:
            return signer.role in (Roles.TRUSTEE, Roles.STEWARD)

    if action_id == ActionIds.add:
        return check_promotion()

    elif action_id == ActionIds.edit:
        # is_self = signer.did == did_ledger.did
        is_owner = signer == (did_ledger if did_ledger.verkey is not None else
        did_ledger.creator)

        if (VERKEY in op) and (not is_owner):
            return False

        if ROLE in op:
            if op_role == did_ledger.role:
                # FIXME INDY-1969: related to the task
                return is_owner  # TODO what is a case here, is it correctly designed

            elif op_role == Roles.IDENTITY_OWNER:  # demotion of existent DID
                return check_demotion()

            elif did_ledger.role == Roles.IDENTITY_OWNER:  # promotion of existent DID
                return check_promotion()

            else:  # role updating: demotion + promotion
                return (check_demotion() and check_promotion())
        else:
            return True

    return False


def sign_and_validate(looper, node, action_id, signer, op, did_ledger=None):
    req_obj = sdk_gen_request(op, protocol_version=CURRENT_PROTOCOL_VERSION,
                              identifier=signer.did)
    s_req = sdk_sign_request_objects(looper, signer.wallet_did, [req_obj])[0]

    request = Request(**json.loads(s_req))

    if auth_check(action_id, signer, op, did_ledger):
        node.write_manager.dynamic_validation(request, 0)
    else:
        with pytest.raises(UnauthorizedClientRequest):
            node.write_manager.dynamic_validation(request, 0)


# TESTS
# Note. some fixtures are referred explicitly just to make test nodeid names predictable

def test_nym_add(
        provisioner_role, nym_add_dest_role, nym_add_dest_verkey,
        looper, txnPoolNodeSet,
        provisioner, add_op):
    sign_and_validate(looper, txnPoolNodeSet[0], ActionIds.add, provisioner, add_op)


def test_nym_edit(
        edited_ledger_role, edited_ledger_verkey, editor_type,
        edited_nym_role, edited_nym_verkey,
        looper, txnPoolNodeSet,
        editor, edited, edit_op):
    if edit_op is None:  # might be None, means a duplicate test case
        return

    if editor.verkey is None:  # skip that as well since it doesn't make sense
        return

    if not ROLE in edit_op:  # skip if the update operation changes neither role nor verkey
        if not VERKEY in edit_op:
            return

    sign_and_validate(looper, txnPoolNodeSet[0], ActionIds.edit, editor, edit_op, did_ledger=edited)



from indy_node.server.upgrade_log import UpgradeLog
from indy_node.test import waits
from indy_node.test.upgrade.helper import checkUpgradeScheduled, sdk_ensure_upgrade_sent
from plenum.common.constants import VERSION
from plenum.common.messages.node_messages import Propagate
from plenum.common.request import Request
from plenum.test.delayers import req_delay, ppgDelay
from plenum.test.test_node import getNonPrimaryReplicas


def test_forced_upgrade_handled_once_if_request_received_after_propagate(
        looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee,
        validUpgradeExpForceTrue):
    """
    Verifies that POOL_UPGRADE force=true request is handled one time in case
    the node commits the transaction to the ledger but during the 3PC-process
    receives the request directly from the client after a PROPAGATE from some
    other node
    """
    slow_node = getNonPrimaryReplicas(nodeSet, instId=0)[-1].node

    slow_node.clientIbStasher.delay(req_delay())
    slow_node.nodeIbStasher.delay(ppgDelay(sender_filter='Beta'))
    slow_node.nodeIbStasher.delay(ppgDelay(sender_filter='Gamma'))

    original_process_propagate = slow_node.nodeMsgRouter.routes[Propagate]
    original_process_request = slow_node.clientMsgRouter.routes[Request]

    def patched_process_propagate(msg: Propagate, frm: str):
        original_process_propagate(msg, frm)
        slow_node.clientIbStasher.reset_delays_and_process_delayeds()
        slow_node.nodeMsgRouter.routes[Propagate] = original_process_propagate

    def patched_process_request(request: Request, frm: str):
        original_process_request(request, frm)
        slow_node.nodeIbStasher.reset_delays_and_process_delayeds()
        slow_node.clientMsgRouter.routes[Request] = original_process_request

    slow_node.nodeMsgRouter.routes[Propagate] = patched_process_propagate
    slow_node.clientMsgRouter.routes[Request] = patched_process_request

    init_len = len(list(slow_node.upgrader._actionLog))

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                            validUpgradeExpForceTrue)

    looper.runFor(waits.expectedUpgradeScheduled())

    checkUpgradeScheduled([slow_node], validUpgradeExpForceTrue[VERSION])
    if init_len ==0:
        # first upgrade - should be only one scheduled
        assert len(list(slow_node.upgrader._actionLog)) == 1
    else:
        # one upgrade were already scheduled. we should cancel it and schedule new one
        # so action log should be increased by 2
        assert len(list(slow_node.upgrader._actionLog)) == init_len + 2
    assert slow_node.upgrader._actionLog.last_event.ev_type == UpgradeLog.Events.scheduled



import pytest as pytest

from common.serializers.serialization import domain_state_serializer
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_aml_handler import \
    TxnAuthorAgreementAmlHandler
from indy_node.test.request_handlers.helper import get_exception, add_to_idr
from plenum.common.constants import ROLE, STEWARD, NYM, TARGET_NYM, TXN_TYPE, TXN_AUTHOR_AGREEMENT, \
    TXN_AUTHOR_AGREEMENT_TEXT, TXN_AUTHOR_AGREEMENT_VERSION, TRUSTEE, DOMAIN_LEDGER_ID, TXN_AUTHOR_AGREEMENT_AML, \
    AML_VERSION, AML, AML_CONTEXT
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.request import Request
from plenum.test.testing_utils import FakeSomething


@pytest.fixture(scope="module")
def txn_author_agreement_aml_handler(db_manager, write_auth_req_validator):
    handler = TxnAuthorAgreementAmlHandler(db_manager, write_auth_req_validator)
    return handler


@pytest.fixture(scope="module")
def aml_request(txn_author_agreement_aml_handler, creator):
    return Request(identifier=creator,
                   signature="signature",
                   operation={TXN_TYPE: TXN_AUTHOR_AGREEMENT_AML,
                              AML_VERSION: "AML_VERSION",
                              AML: {"test": "test"},
                              AML_CONTEXT: "AML_CONTEXT"})


def test_dynamic_validation_without_permission(aml_request,
                                               txn_author_agreement_aml_handler: TxnAuthorAgreementAmlHandler,
                                               creator):
    add_to_idr(txn_author_agreement_aml_handler.database_manager.idr_cache, creator, STEWARD)
    with pytest.raises(UnauthorizedClientRequest, match="Not enough TRUSTEE signatures"):
        txn_author_agreement_aml_handler.dynamic_validation(aml_request, 0)


def test_dynamic_validation(aml_request,
                            txn_author_agreement_aml_handler: TxnAuthorAgreementAmlHandler,
                            creator):
    add_to_idr(txn_author_agreement_aml_handler.database_manager.idr_cache, creator, TRUSTEE)
    txn_author_agreement_aml_handler.dynamic_validation(aml_request, 0)



from collections import OrderedDict
from indy_common.types import ClientGetRevocRegDefField
from plenum.common.messages.fields import NonEmptyStringField, ConstantField
from indy_common.constants import ID, TXN_TYPE, REVOC_TYPE


EXPECTED_GET_REVOC_REG_DEF_FIELD = OrderedDict([
    (ID, NonEmptyStringField()),
    (TXN_TYPE, ConstantField),
])


def test_get_revoc_reg_def_schema():
    actual_field_name = OrderedDict(ClientGetRevocRegDefField.schema).keys()
    assert EXPECTED_GET_REVOC_REG_DEF_FIELD.keys() == actual_field_name



#!/usr/bin/python3.5

from stp_core.common.log import getlogger

logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"
CLIENT_CONNECTIONS_LIMIT = 15360


def migrate_env():
    line = "CLIENT_CONNECTIONS_LIMIT={}".format(CLIENT_CONNECTIONS_LIMIT)
    with open(ENV_FILE_PATH, "a") as envfile:
        logger.info("Append '{}' line to '{}'."
                    .format(line, ENV_FILE_PATH))
        envfile.write("\n" + line)


migrate_env()



import pytest
import time

from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.authorize.authorizer import RolesAuthorizer
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import CURRENT_PROTOCOL_VERSION, TARGET_NYM, TXN_TYPE, NYM, STEWARD, TRUSTEE
from plenum.test.helper import randomOperation
from storage.kv_in_memory import KeyValueStorageInMemory


@pytest.fixture(scope='module')
def req_auth():
    return Request(identifier="some_identifier",
                   reqId=1,
                   operation=randomOperation(),
                   signature="signature",
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


@pytest.fixture(scope='module')
def req_multi_signed_by_non_author():
    return Request(identifier="some_identifier",
                   reqId=1,
                   operation=randomOperation(),
                   signatures={"some_identifier2": "sig"},
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


@pytest.fixture(scope='module')
def idr_cache(req_auth):
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    cache.set(req_auth.identifier, 1, int(time.time()), role=STEWARD,
              verkey="SomeVerkey", isCommitted=False),
    cache.set("some_identifier2", 1, int(time.time()), role=STEWARD,
              verkey="SomeVerkey2", isCommitted=False)
    return cache


@pytest.fixture(scope='module')
def idr_cache_none_role(req_auth):
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    cache.set(req_auth.identifier, 1, int(time.time()),
              verkey="SomeVerkey", isCommitted=False)
    return cache


def test_role_authorizer_get_role(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    assert authorizer.get_role(req_auth) == STEWARD


def test_role_authorizer_is_role_accepted(idr_cache):
    authorizer = RolesAuthorizer(cache=idr_cache)
    assert authorizer.is_role_accepted(role="", auth_constraint_role=IDENTITY_OWNER)
    assert not authorizer.is_role_accepted(role=TRUSTEE, auth_constraint_role=IDENTITY_OWNER)
    assert not authorizer.is_role_accepted(role=None, auth_constraint_role=IDENTITY_OWNER)
    assert authorizer.is_role_accepted(role=TRUSTEE, auth_constraint_role=TRUSTEE)
    assert authorizer.is_role_accepted(role="", auth_constraint_role="*")


def test_role_authorizer_is_owner_accepted(idr_cache, is_owner):
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized = is_owner
    assert authorized == authorizer.is_owner_accepted(
        AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=True),
        AuthActionAdd(txn_type=NYM, field='some_field', value='some_value', is_owner=is_owner))


def test_role_authorizer_authorize_with_owner(idr_cache, req_auth, is_owner):
    req = Request(identifier=req_auth.identifier,
                  operation={TARGET_NYM: req_auth.identifier,
                             TXN_TYPE: NYM},
                  signature='signature')
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized, reason = authorizer.authorize(req,
                                              AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=True),
                                              AuthActionAdd(txn_type=NYM, field='some_field', value='some_value',
                                                            is_owner=is_owner))
    assert authorized == is_owner


def test_role_authorizer_authorize_with_role(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role=STEWARD, sig_count=1))
    assert authorized


def test_role_authorizer_not_authorize_role(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role=TRUSTEE, sig_count=1))
    assert not authorized
    assert reason == "Not enough TRUSTEE signatures"


def test_role_authorizer_not_authorize_unknown_nym(idr_cache):
    authorizer = RolesAuthorizer(cache=idr_cache)

    unknown_req_auth = Request(identifier="some_unknown_identifier",
                               reqId=2,
                               operation=randomOperation(),
                               signature="signature",
                               protocolVersion=CURRENT_PROTOCOL_VERSION)

    authorized, reason = authorizer.authorize(unknown_req_auth,
                                              AuthConstraint(role=TRUSTEE, sig_count=1))
    assert not authorized
    assert reason == "sender's DID {} is not found in the Ledger".format(unknown_req_auth.identifier)


def test_role_authorizer_is_sig_count_accepted(idr_cache_none_role, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache_none_role)
    assert authorizer.is_sig_count_accepted(req_auth, AuthConstraint(role="*", sig_count=1))


def test_role_authorizer_not_is_sig_count_accepted(idr_cache_none_role, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache_none_role)
    assert not authorizer.is_sig_count_accepted(req_auth, AuthConstraint(role=TRUSTEE, sig_count=10))


def test_role_authorizer_off_ledger_signature_pass(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    req_auth._identifier = 'id_off_ledger'
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role='*', sig_count=1,
                                                                       off_ledger_signature=True))
    assert authorized


def test_role_authorizer_off_ledger_signature_not_pass(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    req_auth._identifier = 'id_off_ledger'
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role='*', sig_count=1,
                                                                       off_ledger_signature=False))
    assert not authorized
    assert "DID id_off_ledger is not found in the Ledger" in reason


def test_role_authorizer_off_ledger_signature_count_2_pass(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    req_auth._identifier = 'id_off_ledger'
    req_auth.signature = None
    req_auth.signatures = {'id_off_ledger': 'signature', 'another_id_off_ledger': 'another_signature'}
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role='*', sig_count=2,
                                                                       off_ledger_signature=True))
    assert authorized


def test_role_authorizer_off_ledger_signature_count_2_different_pass(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    req_auth.signature = None
    req_auth.signatures = {req_auth.identifier: 'signature', 'another_id_off_ledger': 'another_signature'}
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role='*', sig_count=2,
                                                                       off_ledger_signature=True))
    assert authorized


def test_role_authorizer_off_ledger_signature_count_0_pass(idr_cache, req_auth):
    authorizer = RolesAuthorizer(cache=idr_cache)
    req_auth._identifier = 'id_off_ledger'
    authorized, reason = authorizer.authorize(req_auth, AuthConstraint(role='*', sig_count=0,
                                                                       off_ledger_signature=True))
    assert authorized


def test_signed_by_author_if_more_than_1_sig(idr_cache, req_multi_signed_by_non_author):
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized, reason = authorizer.authorize(req_multi_signed_by_non_author,
                                              AuthConstraint(role=STEWARD, sig_count=1))
    assert not authorized
    assert "Author must sign the transaction" in reason


def test_no_sign_by_author_if_0_sig(idr_cache, req_multi_signed_by_non_author):
    authorizer = RolesAuthorizer(cache=idr_cache)
    authorized, reason = authorizer.authorize(req_multi_signed_by_non_author,
                                              AuthConstraint(role=STEWARD, sig_count=0))
    assert authorized



import json

import pytest
from indy.ledger import build_acceptance_mechanisms_request

from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_map import steward_or_trustee_constraint
from indy_node.test.helper import sdk_send_and_check_auth_rule_request
from plenum.common.constants import TXN_AUTHOR_AGREEMENT_VERSION, TXN_AUTHOR_AGREEMENT_TEXT, \
    TXN_AUTHOR_AGREEMENT_RATIFICATION_TS, TXN_AUTHOR_AGREEMENT, REPLY
from plenum.common.exceptions import RequestRejectedException
from plenum.common.util import randomString, get_utc_epoch
from plenum.test.txn_author_agreement.helper import sdk_send_txn_author_agreement, sdk_get_txn_author_agreement


def test_send_valid_txn_author_agreement_succeeds(looper, setup_aml, txnPoolNodeSet, sdk_pool_handle,
                                                  sdk_wallet_trustee, sdk_wallet_client):
    text = randomString(1024)
    version = randomString(16)
    ratified = get_utc_epoch() - 300
    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version, text, ratified=ratified)

    reply = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_client)[1]
    assert reply['op'] == REPLY

    result = reply['result']['data']
    assert result[TXN_AUTHOR_AGREEMENT_TEXT] == text
    assert result[TXN_AUTHOR_AGREEMENT_VERSION] == version
    assert result[TXN_AUTHOR_AGREEMENT_RATIFICATION_TS] == ratified


def test_send_valid_txn_author_agreement_without_enough_privileges_fails(looper, setup_aml, txnPoolNodeSet,
                                                                         sdk_pool_handle,
                                                                         sdk_wallet_steward):
    with pytest.raises(RequestRejectedException):
        sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_steward,
                                      randomString(16), randomString(1024))


def test_txn_author_agreement_respects_current_auth_rules(looper, setup_aml, txnPoolNodeSet, sdk_pool_handle,
                                                          sdk_wallet_trustee, sdk_wallet_steward):
    sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX, auth_type=TXN_AUTHOR_AGREEMENT,
                                         field='*', new_value='*',
                                         constraint=steward_or_trustee_constraint.as_dict)

    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_steward,
                                  randomString(16), randomString(1024), ratified=get_utc_epoch() - 300)



from typing import Optional

from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from plenum.common.constants import LEDGERS_FREEZE
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.ledgers_freeze.ledgers_freeze_handler import LedgersFreezeHandler as PLedgersFreezeHandler


class LedgersFreezeHandler(PLedgersFreezeHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager)
        self.write_req_validator = write_req_validator

    def authorize(self, request):
        self.write_req_validator.validate(request,
                                          [AuthActionEdit(txn_type=LEDGERS_FREEZE,
                                                          field='*',
                                                          old_value='*',
                                                          new_value='*')])



from abc import abstractmethod, ABCMeta

from indy_common.authorize.auth_actions import split_action_id
from indy_common.authorize.auth_constraints import AbstractAuthConstraint, AbstractConstraintSerializer
from indy_common.state import config
from plenum.common.metrics_collector import MetricsName, MetricsCollector
from state.pruning_state import PruningState
from stp_core.common.log import getlogger

logger = getlogger()


class AbstractAuthStrategy(metaclass=ABCMeta):
    def __init__(self, auth_map):
        self.auth_map = auth_map

    @abstractmethod
    def get_auth_constraint(self, action_id) -> AbstractAuthConstraint:
        raise NotImplementedError()

    @abstractmethod
    def _find_auth_constraint_key(self, action_id, auth_map):
        raise NotImplementedError()

    @staticmethod
    def is_accepted_action_id(from_auth_map, from_req):
        am = split_action_id(from_auth_map)
        r = split_action_id(from_req)
        if r.prefix != am.prefix:
            return False
        if r.txn_type != am.txn_type:
            return False
        if r.field != am.field and \
                am.field != '*':
            return False
        if r.old_value != am.old_value and \
                am.old_value != '*':
            return False
        if r.new_value != am.new_value and \
                am.new_value != '*':
            return False
        return True


class LocalAuthStrategy(AbstractAuthStrategy):

    def get_auth_constraint(self, action_id) -> AbstractAuthConstraint:
        am_id = self._find_auth_constraint_key(action_id, self.auth_map)
        return self.auth_map.get(am_id)

    def _find_auth_constraint_key(self, action_id, auth_map):
        for am_id in auth_map.keys():
            if self.is_accepted_action_id(am_id, action_id):
                return am_id


class ConfigLedgerAuthStrategy(AbstractAuthStrategy):

    def __init__(self,
                 auth_map,
                 state: PruningState,
                 serializer: AbstractConstraintSerializer,
                 metrics: MetricsCollector = None):
        super().__init__(auth_map=auth_map)
        self.state = state
        self.serializer = serializer
        self.metrics = metrics
        self.from_state_count = 0

    def get_auth_constraint(self, action_id: str) -> AbstractAuthConstraint:
        """
        Find rule_id for incoming action_id and return AuthConstraint instance
        """
        return self._find_auth_constraint(action_id, self.auth_map)

    def _find_auth_constraint(self, action_id, auth_map):
        am_id = self._find_auth_constraint_key(action_id, auth_map)
        if am_id:
            constraint = self.get_from_state(key=config.make_state_path_for_auth_rule(am_id))
            if not constraint:
                return auth_map.get(am_id)
            logger.debug("Using auth constraint from state")
            if self.metrics:
                self.from_state_count += 1
                self.metrics.add_event(MetricsName.AUTH_RULES_FROM_STATE_COUNT, self.from_state_count)
            return constraint

    def _find_auth_constraint_key(self, action_id, auth_map):
        for am_id in auth_map.keys():
            if self.is_accepted_action_id(am_id, action_id):
                return am_id

    def get_from_state(self, key, isCommitted=False):
        from_state = self.state.get(key=key,
                                    isCommitted=isCommitted)
        if not from_state:
            return None
        return self.serializer.deserialize(from_state)



import json
from contextlib import contextmanager

import base58
from indy.did import replace_keys_start, replace_keys_apply
from indy.ledger import (
    build_attrib_request, build_get_attrib_request,
    build_get_auth_rule_request, build_auth_rule_request,
    build_auth_rules_request)
from libnacl import randombytes

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import ROLE, CONSTRAINT_ID, ConstraintsEnum, SIG_COUNT, NEED_TO_BE_OWNER, \
    METADATA, OFF_LEDGER_SIGNATURE
from indy_common.config_helper import NodeConfigHelper
from indy_common.constants import NYM, ENDORSER, CONSTRAINT, AUTH_ACTION, AUTH_TYPE, FIELD, NEW_VALUE, OLD_VALUE
from indy_common.test.helper import TempStorage
from indy_node.server.node import Node
from indy_node.server.node_bootstrap import NodeBootstrap
from indy_node.server.upgrader import Upgrader
from plenum.common.constants import TRUSTEE
from plenum.common.signer_did import DidSigner
from plenum.common.signer_simple import SimpleSigner
from plenum.common.util import rawToFriendly
from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_req
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.test_node import TestNodeCore
from plenum.test.testable import spyable
from stp_core.common.log import getlogger
from stp_core.types import HA

logger = getlogger()


@spyable(methods=[Upgrader.processLedger])
class TestUpgrader(Upgrader):
    pass


class TestNodeBootstrap(NodeBootstrap):
    def init_upgrader(self):
        return TestUpgrader(self.node.id, self.node.name, self.node.dataLocation, self.node.config,
                            self.node.configLedger,
                            actionFailedCallback=self.node.postConfigLedgerCaughtUp,
                            action_start_callback=self.node.notify_upgrade_start)


# noinspection PyShadowingNames,PyShadowingNames
@spyable(
    methods=[Node.handleOneNodeMsg, Node.processRequest, Node.processOrdered,
             Node.postToClientInBox, Node.postToNodeInBox, "eatTestMsg",
             Node.discard,
             Node.reportSuspiciousNode, Node.reportSuspiciousClient,
             Node.processRequest, Node.processPropagate, Node.propagate,
             Node.forward, Node.send, Node.checkPerformance,
             Node.getReplyFromLedger, Node.getReplyFromLedgerForRequest,
             Node.no_more_catchups_needed, Node.onBatchCreated,
             Node.onBatchRejected])
class TestNode(TempStorage, TestNodeCore, Node):
    def __init__(self, *args, **kwargs):
        from plenum.common.stacks import nodeStackClass, clientStackClass
        self.NodeStackClass = nodeStackClass
        self.ClientStackClass = clientStackClass
        if kwargs.get('bootstrap_cls', None) is None:
            kwargs['bootstrap_cls'] = TestNodeBootstrap

        Node.__init__(self, *args, **kwargs)
        TestNodeCore.__init__(self, *args, **kwargs)
        self.cleanupOnStopping = True

    def init_domain_req_handler(self):
        return Node.init_domain_req_handler(self)

    def init_core_authenticator(self):
        return Node.init_core_authenticator(self)

    def onStopping(self, *args, **kwargs):
        # self.graphStore.store.close()
        super().onStopping(*args, **kwargs)
        if self.cleanupOnStopping:
            self.cleanupDataLocation()

    def schedule_node_status_dump(self):
        pass

    def dump_additional_info(self):
        pass

    @property
    def nodeStackClass(self):
        return self.NodeStackClass

    @property
    def clientStackClass(self):
        return self.ClientStackClass


# TODO makes sense to move plenum
def sdk_send_and_check_req_json(
    looper, sdk_pool_handle, sdk_wallet, req_json, no_wait=False
):
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet, req_json)
    if no_wait:
        return req
    resp = sdk_get_and_check_replies(looper, [req])
    return resp


def sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_handle, attrib,
                                dest=None, xhash=None, enc=None):
    _, s_did = sdk_wallet_handle
    t_did = dest or s_did
    attrib_req = looper.loop.run_until_complete(
        build_attrib_request(s_did, t_did, xhash, attrib, enc))
    return sdk_send_and_check_req_json(looper, sdk_pool_handle, sdk_wallet_handle, attrib_req)


def sdk_get_attribute_and_check(looper, sdk_pool_handle, submitter_wallet, target_did, attrib_name):
    _, submitter_did = submitter_wallet
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, target_did, attrib_name, None, None))
    return sdk_send_and_check_req_json(looper, sdk_pool_handle, submitter_wallet, req)


def sdk_add_raw_attribute(looper, sdk_pool_handle, sdk_wallet_handle, name, value):
    _, did = sdk_wallet_handle
    attrData = json.dumps({name: value})
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_handle, attrData)


# AUTH_RULE / GET_AUTH_RULE helpers


def build_auth_rule_request_json(
    looper, submitter_did,
    auth_action, auth_type, field, constraint, old_value=None, new_value=None
):
    return looper.loop.run_until_complete(
        build_auth_rule_request(
            submitter_did=submitter_did,
            txn_type=auth_type,
            action=auth_action,
            field=field,
            old_value=old_value,
            new_value=new_value,
            constraint=json.dumps(constraint)
        )
    )


def build_get_auth_rule_request_json(
    looper, submitter_did,
    auth_type=None,
    auth_action=None,
    field=None,
    old_value=None,
    new_value=None
):
    return looper.loop.run_until_complete(
        build_get_auth_rule_request(
            submitter_did=submitter_did,
            txn_type=auth_type,
            action=auth_action,
            field=field,
            old_value=old_value,
            new_value=new_value
        )
    )


def sdk_send_and_check_auth_rule_request(
    looper, sdk_pool_handle, sdk_wallet,
    auth_action, auth_type, field, constraint, old_value=None, new_value=None,
    no_wait=False
):
    req_json = build_auth_rule_request_json(
        looper, sdk_wallet[1],
        auth_action=auth_action,
        auth_type=auth_type,
        field=field,
        old_value=old_value,
        new_value=new_value,
        constraint=constraint
    )

    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet, req_json, no_wait=no_wait
    )


def sdk_send_and_check_auth_rules_request(looper, sdk_pool_handle,
                                          sdk_wallet, rules=None, no_wait=False):
    if rules is None:
        rules = [generate_auth_rule(ADD_PREFIX, NYM,
                                    ROLE, ENDORSER),
                 generate_auth_rule(EDIT_PREFIX, NYM,
                                    ROLE, ENDORSER, TRUSTEE)]
    req_json = looper.loop.run_until_complete(
        build_auth_rules_request(submitter_did=sdk_wallet[1], data=json.dumps(rules)))
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet, req_json, no_wait=no_wait
    )


def sdk_send_and_check_get_auth_rule_request(
    looper, sdk_pool_handle, sdk_wallet,
    auth_type=None,
    auth_action=None,
    field=None,
    old_value=None,
    new_value=None
):
    req_json = build_get_auth_rule_request_json(
        looper, sdk_wallet[1],
        auth_type=auth_type,
        auth_action=auth_action,
        field=field,
        old_value=old_value,
        new_value=new_value
    )
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet, req_json
    )


def generate_auth_rule(auth_action=ADD_PREFIX, auth_type=NYM,
                       field=ROLE, new_value=ENDORSER,
                       old_value=None, constraint=None):
    if constraint is None:
        constraint = generate_constraint_entity()
    rule = {CONSTRAINT: constraint,
            AUTH_ACTION: auth_action,
            AUTH_TYPE: auth_type,
            FIELD: field,
            NEW_VALUE: new_value
            }
    if old_value or auth_action == EDIT_PREFIX:
        rule[OLD_VALUE] = old_value
    return rule


def generate_constraint_entity(constraint_id=ConstraintsEnum.ROLE_CONSTRAINT_ID,
                               role=TRUSTEE,
                               sig_count=1,
                               need_to_be_owner=False,
                               off_ledger_signature=None,
                               metadata={}):
    constraint = {CONSTRAINT_ID: constraint_id,
                  ROLE: role,
                  SIG_COUNT: sig_count,
                  NEED_TO_BE_OWNER: need_to_be_owner,
                  METADATA: metadata}
    if off_ledger_signature is not None:
        constraint[OFF_LEDGER_SIGNATURE] = off_ledger_signature
    return constraint


base58_alphabet = set(base58.alphabet.decode("utf-8"))


def check_str_is_base58_compatible(str):
    return not (set(str) - base58_alphabet)


def sdk_rotate_verkey(looper, sdk_pool_handle, wh,
                      did_of_changer,
                      did_of_changed,
                      verkey=None):
    verkey = verkey or looper.loop.run_until_complete(
        replace_keys_start(wh, did_of_changed, json.dumps({})))

    sdk_add_new_nym(looper, sdk_pool_handle,
                    (wh, did_of_changer), dest=did_of_changed,
                    verkey=verkey)
    looper.loop.run_until_complete(
        replace_keys_apply(wh, did_of_changed))
    return verkey


def start_stopped_node(stopped_node, looper, tconf, tdir, allPluginsPath, start=True):
    nodeHa, nodeCHa = HA(*
                         stopped_node.nodestack.ha), HA(*
                                                        stopped_node.clientstack.ha)
    config_helper = NodeConfigHelper(stopped_node.name, tconf, chroot=tdir)
    restarted_node = TestNode(stopped_node.name,
                              config_helper=config_helper,
                              config=tconf,
                              ha=nodeHa, cliha=nodeCHa,
                              pluginPaths=allPluginsPath)
    if start:
        looper.add(restarted_node)
    return restarted_node


def modify_field(string, value, *field_path):
    d = json.loads(string)
    prev = None
    for i in range(0, len(field_path) - 1):
        if prev is None:
            prev = d[field_path[i]]
            continue
        prev = prev[field_path[i]]
    if prev:
        prev[field_path[-1]] = value
    else:
        d[field_path[-1]] = value
    return json.dumps(d)


def createUuidIdentifier():
    return rawToFriendly(randombytes(16))


def createHalfKeyIdentifierAndAbbrevVerkey(seed=None):
    didSigner = DidSigner(seed=seed)
    return didSigner.identifier, didSigner.verkey


def createCryptonym(seed=None):
    return SimpleSigner(seed=seed).identifier


def createUuidIdentifierAndFullVerkey(seed=None):
    didSigner = DidSigner(identifier=createUuidIdentifier(), seed=seed)
    return didSigner.identifier, didSigner.verkey


@contextmanager
def rich_schemas_enabled_scope(tconf):
    old_value = tconf.ENABLE_RICH_SCHEMAS
    tconf.ENABLE_RICH_SCHEMAS = True
    yield tconf
    tconf.ENABLE_RICH_SCHEMAS = old_value



import os
import asyncio
from datetime import datetime
from functools import partial
from typing import Optional, Callable, Dict

import dateutil.parser
import dateutil.tz

from indy_node.server.node_maintainer import NodeMaintainer, \
    NodeControlToolMessage
from plenum.common.txn_util import is_forced, get_seq_no, get_type, get_payload_data, get_req_id, get_from
from stp_core.common.log import getlogger
from plenum.common.constants import VERSION
from common.version import (
    SourceVersion, InvalidVersionError
)

from indy_common.constants import ACTION, POOL_UPGRADE, START, SCHEDULE, \
    CANCEL, JUSTIFICATION, TIMEOUT, NODE_UPGRADE, \
    UPGRADE_MESSAGE, PACKAGE, APP_NAME
from indy_common.version import src_version_cls
from indy_node.server.upgrade_log import UpgradeLogData, UpgradeLog
from indy_node.utils.node_control_utils import NodeControlUtil

logger = getlogger()


class Upgrader(NodeMaintainer):

    def __init__(self, nodeId, nodeName, dataDir, config, ledger=None,
                 actionLog=None, actionFailedCallback: Callable = None,
                 action_start_callback: Callable = None):
        self._should_notify_about_upgrade = False
        super().__init__(nodeId, nodeName, dataDir, config, ledger, actionLog,
                         actionFailedCallback, action_start_callback)

    @staticmethod
    def get_src_version(
            pkg_name: str = APP_NAME,
            nocache: bool = False) -> SourceVersion:

        if pkg_name == APP_NAME and not nocache:
            from indy_node.__metadata__ import __version__
            return src_version_cls(APP_NAME)(__version__)

        curr_pkg_ver, _ = NodeControlUtil.curr_pkg_info(pkg_name)
        return curr_pkg_ver.upstream if curr_pkg_ver else None

    @staticmethod
    def is_version_upgradable(
            old: SourceVersion, new: SourceVersion, reinstall: bool = False):
        return (new > old) or (new == old and reinstall)

    @staticmethod
    def get_action_id(txn):
        seq_no = get_seq_no(txn) or ''
        if is_forced(txn):
            seq_no = ''
        return '{}{}'.format(get_req_id(txn), seq_no)

    # implements legacy logic used only by migration logic,
    # please use Version classes in code instead
    # TODO refactor migration logic to get rid of that API usage
    @staticmethod
    def compareVersions(verA: str, verB: str) -> int:
        version_cls = src_version_cls(APP_NAME)
        return version_cls.cmp(version_cls(verA), version_cls(verB))

    def _defaultLog(self, dataDir, config):
        log = os.path.join(dataDir, config.upgradeLogFile)
        return UpgradeLog(file_path=log)

    def _is_action_started(self):
        last_action = self.lastActionEventInfo
        if not last_action:
            logger.debug('Node {} has no upgrade events'.format(self.nodeName))
            return False

        if last_action.ev_type != UpgradeLog.Events.started:
            logger.debug(
                "Upgrade for node {} was not scheduled. "
                "Last event is {}"
                .format(self.nodeName, last_action))
            return False

        return True

    def _update_action_log_for_started_action(self):
        ev_data = self.lastActionEventInfo.data

        self._should_notify_about_upgrade = True
        if not self.didLastExecutedUpgradeSucceeded:
            self._actionLog.append_failed(ev_data)
            self._action_failed(ev_data, external_reason=True)
            return

        self._actionLog.append_succeeded(ev_data)
        logger.info(
            "Node '{}' successfully upgraded to version {}"
            .format(self.nodeName, ev_data.version))
        self._notifier.sendMessageUponNodeUpgradeComplete(
            "Upgrade of package {} on node '{}' to version {} scheduled on {} "
            " with upgrade_id {} completed successfully"
            .format(ev_data.pkg_name, self.nodeName,
                    ev_data.version, ev_data.when, ev_data.upgrade_id))

    def should_notify_about_upgrade_result(self):
        # do not rely on NODE_UPGRADE txn in config ledger, since in
        # some cases (for example, when
        # we run POOL_UPGRADE with force=true), we may not have
        # IN_PROGRESS NODE_UPGRADE in the ledger.

        # send NODE_UPGRADE txn only if we were in Upgrade Started
        # state at the very beginning (after Node restarted)
        return self._should_notify_about_upgrade

    def notified_about_action_result(self):
        self._should_notify_about_upgrade = False

    def get_last_node_upgrade_txn(self, start_no: int = None):
        return self.get_upgrade_txn(
            lambda txn: get_type(txn) == NODE_UPGRADE and get_from(txn) == self.nodeId,
            start_no=start_no,
            reverse=True)

    def get_upgrade_txn(self, predicate: Callable = None, start_no: int = None,
                        reverse: bool = False) -> Optional[Dict]:
        def txn_filter(txn):
            return not predicate or predicate(txn)

        def traverse_end_condition(seq_no):
            if reverse:
                return seq_no > 0
            return seq_no <= len(self.ledger)

        inc = 1
        init_start_no = 1
        if reverse:
            inc = -1
            init_start_no = len(self.ledger)

        seq_no = start_no if start_no is not None else init_start_no
        while traverse_end_condition(seq_no):
            txn = self.ledger.getBySeqNo(seq_no)
            if txn_filter(txn):
                return txn
            seq_no += inc
        return None

    # TODO: PoolConfig and Updater both read config ledger independently
    def processLedger(self) -> None:
        """
        Checks ledger for planned but not yet performed upgrades
        and schedules upgrade for the most recent one

        Assumption: Only version is enough to identify a release, no hash
        checking is done
        :return:
        """
        logger.debug(
            '{} processing config ledger for any upgrades'.format(self))
        last_pool_upgrade_txn_start = self.get_upgrade_txn(
            lambda txn: get_type(txn) == POOL_UPGRADE and get_payload_data(txn)[ACTION] == START, reverse=True)
        if last_pool_upgrade_txn_start:
            logger.info('{} found upgrade START txn {}'.format(
                self, last_pool_upgrade_txn_start))
            last_pool_upgrade_txn_seq_no = get_seq_no(last_pool_upgrade_txn_start)

            # searching for CANCEL for this upgrade submitted after START txn
            last_pool_upgrade_txn_cancel = self.get_upgrade_txn(
                lambda txn:
                get_type(txn) == POOL_UPGRADE and get_payload_data(txn)[ACTION] == CANCEL and get_payload_data(txn)
                [VERSION] == get_payload_data(last_pool_upgrade_txn_start)[VERSION],
                start_no=last_pool_upgrade_txn_seq_no + 1)
            if last_pool_upgrade_txn_cancel:
                logger.info('{} found upgrade CANCEL txn {}'.format(
                    self, last_pool_upgrade_txn_cancel))
                return

            self.handleUpgradeTxn(last_pool_upgrade_txn_start)

    # TODO might necessary to improve tests
    @property
    def didLastExecutedUpgradeSucceeded(self) -> bool:
        """
        Checks last record in upgrade log to find out whether it
        is about scheduling upgrade. If so - checks whether current version
        is equals to the one in that record

        :returns: upgrade execution result
        """
        lastEventInfo = self.lastActionEventInfo
        if lastEventInfo:
            ev_data = lastEventInfo.data
            currentPkgVersion = NodeControlUtil.curr_pkg_info(ev_data.pkg_name)[0]
            if currentPkgVersion:
                return currentPkgVersion.upstream == ev_data.version
            else:
                logger.warning(
                    "{} failed to get information about package {} "
                    "scheduled for last upgrade"
                    .format(self, ev_data.pkg_name)
                )
        return False

    @staticmethod
    def check_upgrade_possible(
            pkg_name: str,
            target_ver: str,
            reinstall: bool = False
    ):
        version_cls = src_version_cls(pkg_name)

        try:
            target_ver = version_cls(target_ver)
        except InvalidVersionError:
            return (
                "invalid target version {} for version class {}: "
                .format(target_ver, version_cls)
            )

        # get current installed package version of pkg_name
        curr_pkg_ver, cur_deps = NodeControlUtil.curr_pkg_info(pkg_name)
        if not curr_pkg_ver:
            return ("package {} is not installed and cannot be upgraded"
                    .format(pkg_name))

        # TODO weak check
        if APP_NAME not in pkg_name and all([APP_NAME not in d for d in cur_deps]):
            return "Package {} doesn't belong to pool".format(pkg_name)

        # compare whether it makes sense to try (target >= current, = for reinstall)
        if not Upgrader.is_version_upgradable(
                curr_pkg_ver.upstream, target_ver, reinstall):
            return "Version {} is not upgradable".format(target_ver)

        # get the most recent version of the package for provided version
        # TODO request to NodeControlTool since Node likely runs under user
        # which doesn't have rights to update list of system packages available
        # target_pkg_ver = NodeControlUtil.get_latest_pkg_version(
        #    pkg_name, upstream=target_ver)

        # if not target_pkg_ver:
        #    return ("package {} for target version {} is not found"
        #            .format(pkg_name, target_ver))

        return None

    def handleUpgradeTxn(self, txn) -> None:
        """
        Handles transaction of type POOL_UPGRADE
        Can schedule or cancel upgrade to a newer
        version at specified time

        :param txn:
        """
        FINALIZING_EVENT_TYPES = [UpgradeLog.Events.succeeded, UpgradeLog.Events.failed]

        if get_type(txn) != POOL_UPGRADE:
            return

        logger.info("Node '{}' handles upgrade txn {}".format(self.nodeName, txn))
        txn_data = get_payload_data(txn)
        action = txn_data[ACTION]
        version = txn_data[VERSION]
        justification = txn_data.get(JUSTIFICATION)
        pkg_name = txn_data.get(PACKAGE, self.config.UPGRADE_ENTRY)
        upgrade_id = self.get_action_id(txn)

        # TODO test
        try:
            version = src_version_cls(pkg_name)(version)
        except InvalidVersionError as exc:
            logger.warning(
                "{} can't handle upgrade txn with version {} for package {}: {}"
                .format(self, version, pkg_name, exc)
            )
            return

        if action == START:
            # forced txn could have partial schedule list
            if self.nodeId not in txn_data[SCHEDULE]:
                logger.info("Node '{}' disregards upgrade txn {}".format(
                    self.nodeName, txn))
                return

            last_event = self.lastActionEventInfo
            if last_event:
                if last_event.data.upgrade_id == upgrade_id and last_event.ev_type in FINALIZING_EVENT_TYPES:
                    logger.info(
                        "Node '{}' has already performed an upgrade with upgrade_id {}. "
                        "Last recorded event is {}"
                        .format(self.nodeName, upgrade_id, last_event.data))
                    return

            when = txn_data[SCHEDULE][self.nodeId]
            failTimeout = txn_data.get(TIMEOUT, self.defaultActionTimeout)

            if isinstance(when, str):
                when = dateutil.parser.parse(when)

            new_ev_data = UpgradeLogData(when, version, upgrade_id, pkg_name)

            if self.scheduledAction:
                if self.scheduledAction == new_ev_data:
                    logger.debug(
                        "Node {} already scheduled upgrade to version '{}' "
                        .format(self.nodeName, version))
                    return
                else:
                    logger.info(
                        "Node '{}' cancels previous upgrade and schedules a new one to {}"
                        .format(self.nodeName, version))
                    self._cancelScheduledUpgrade(justification)

            logger.info("Node '{}' schedules upgrade to {}".format(self.nodeName, version))

            self._scheduleUpgrade(new_ev_data, failTimeout)
            return

        if action == CANCEL:
            if self.scheduledAction and self.scheduledAction.version == version:
                self._cancelScheduledUpgrade(justification)
                logger.info("Node '{}' cancels upgrade to {}".format(
                    self.nodeName, version))
            return

        logger.error(
            "Got {} transaction with unsupported action {}".format(
                POOL_UPGRADE, action))

    def _scheduleUpgrade(self,
                         ev_data: UpgradeLogData,
                         failTimeout) -> None:
        """
        Schedules node upgrade to a newer version

        :param ev_data: upgrade event parameters
        """
        logger.info(
            "{}'s upgrader processing upgrade for version {}={}"
            .format(self, ev_data.pkg_name, ev_data.version))
        now = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())

        self._notifier.sendMessageUponNodeUpgradeScheduled(
            "Upgrade of package {} on node '{}' to version {} "
            "has been scheduled on {}"
            .format(ev_data.pkg_name, self.nodeName,
                    ev_data.version, ev_data.when))
        self._actionLog.append_scheduled(ev_data)

        callAgent = partial(self._callUpgradeAgent, ev_data, failTimeout)
        delay = 0
        if now < ev_data.when:
            delay = (ev_data.when - now).total_seconds()
        self.scheduledAction = ev_data
        self._schedule(callAgent, delay)

    def _cancelScheduledUpgrade(self, justification=None) -> None:
        """
        Cancels scheduled upgrade

        :param when: time upgrade was scheduled to
        :param version: version upgrade scheduled for
        """

        if self.scheduledAction:
            why_prefix = ": "
            why = justification
            if justification is None:
                why_prefix = ", "
                why = "cancellation reason not specified"

            ev_data = self.scheduledAction
            logger.info("Cancelling upgrade {}"
                        " of node {}"
                        " of package {}"
                        " to version {}"
                        " scheduled on {}"
                        "{}{}"
                        .format(ev_data.upgrade_id,
                                self.nodeName,
                                ev_data.pkg_name,
                                ev_data.version,
                                ev_data.when,
                                why_prefix,
                                why))

            self._unscheduleAction()
            self._actionLog.append_cancelled(ev_data)
            self._notifier.sendMessageUponPoolUpgradeCancel(
                "Upgrade of package {} on node '{}' to version {} "
                "has been cancelled due to {}"
                .format(ev_data.pkg_name, self.nodeName,
                        ev_data.version, why))

    def _callUpgradeAgent(self, ev_data, failTimeout) -> None:
        """
        Callback which is called when upgrade time come.
        Writes upgrade record to upgrade log and asks
        node control service to perform upgrade

        :param when: upgrade time
        :param version: version to upgrade to
        """

        logger.info("{}'s upgrader calling agent for upgrade".format(self))
        self._actionLog.append_started(ev_data)
        self._action_start_callback()
        self.scheduledAction = None
        asyncio.ensure_future(
            self._sendUpgradeRequest(ev_data, failTimeout))

    async def _sendUpgradeRequest(self, ev_data, failTimeout):
        retryLimit = self.retry_limit
        while retryLimit:
            try:
                msg = UpgradeMessage(
                    version=ev_data.version.full,
                    pkg_name=ev_data.pkg_name
                ).toJson()
                logger.info("Sending message to control tool: {}".format(msg))
                await self._open_connection_and_send(msg)
                break
            except Exception as ex:
                logger.warning("Failed to communicate to control tool: {}".format(ex))
                asyncio.sleep(self.retry_timeout)
                retryLimit -= 1
        if not retryLimit:
            self._action_failed(
                ev_data,
                reason="problems in communication with node control service")
            self._unscheduleAction()
        else:
            logger.info("Waiting {} minutes for upgrade to be performed".format(failTimeout))
            timesUp = partial(self._declareTimeoutExceeded, ev_data)
            self._schedule(timesUp, self.get_timeout(failTimeout))

    def _declareTimeoutExceeded(self, ev_data: UpgradeLogData):
        """
        This function is called when time for upgrade is up
        """
        logger.info("Timeout exceeded for {}:{}"
                    .format(ev_data.when, ev_data.version))
        last = self._actionLog.last_event
        # TODO test this
        if last and last.ev_type == UpgradeLog.Events.failed and last.data == ev_data:
            return None

        self._action_failed(ev_data, reason="exceeded upgrade timeout")

        self._unscheduleAction()
        self._actionFailedCallback()

    def _action_failed(self,
                       ev_data: UpgradeLogData,
                       reason=None,
                       external_reason=False):
        if reason is None:
            reason = "unknown reason"
        error_message = (
            "Node {} failed upgrade {} to "
            "version {} of package {} "
            "scheduled on {} because of {}"
            .format(self.nodeName,
                    ev_data.upgrade_id,
                    ev_data.version,
                    ev_data.pkg_name,
                    ev_data.when,
                    reason)
        )
        logger.error(error_message)
        if external_reason:
            logger.error("This problem may have external reasons, "
                         "check syslog for more information")
        self._notifier.sendMessageUponNodeUpgradeFail(error_message)

    def isScheduleValid(self, schedule, node_srvs, force) -> (bool, str):
        """
        Validates schedule of planned node upgrades

        :param schedule: dictionary of node ids and upgrade times
        :param node_srvs: dictionary of node ids and services
        :return: a 2-tuple of whether schedule valid or not and the reason
        """

        # flag "force=True" ignore basic checks! only datetime format is
        # checked
        times = []
        non_demoted_nodes = set([k for k, v in node_srvs.items() if v])
        if not force and set(schedule.keys()) != non_demoted_nodes:
            return False, 'Schedule should contain id of all nodes'
        now = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
        for dateStr in schedule.values():
            try:
                when = dateutil.parser.parse(dateStr)
                if when <= now and not force:
                    return False, '{} is less than current time'.format(when)
                times.append(when)
            except ValueError:
                return False, '{} cannot be parsed to a time'.format(dateStr)
        if force:
            return True, ''
        times = sorted(times)
        for i in range(len(times) - 1):
            diff = (times[i + 1] - times[i]).total_seconds()
            if diff < self.config.MinSepBetweenNodeUpgrades:
                return False, 'time span between upgrades is {} ' \
                              'seconds which is less than specified ' \
                              'in the config'.format(diff)
        return True, ''


class UpgradeMessage(NodeControlToolMessage):
    """
    Data structure that represents request for node update
    """

    def __init__(self, version: str, pkg_name: str):
        super().__init__(UPGRADE_MESSAGE)
        self.version = version
        self.pkg_name = pkg_name

    def toJson(self):
        import json
        return json.dumps(self.__dict__)



import json
import pytest

from plenum.common.constants import NODE_IP, NODE_PORT, CLIENT_IP, CLIENT_PORT, ALIAS, VALIDATOR, SERVICES
from plenum.common.util import cryptonymToHex, hexToFriendly
from plenum.common.exceptions import RequestNackedException, RequestRejectedException

from plenum.test.helper import sdk_get_and_check_replies, sdk_get_bad_response, sdk_sign_request_strings, \
    sdk_send_signed_requests
from plenum.test.pool_transactions.helper import sdk_add_new_nym, prepare_node_request, \
    sdk_sign_and_send_prepared_request


@pytest.fixture(scope='function')
def node_request(looper, sdk_node_theta_added):
    sdk_steward_wallet, node = sdk_node_theta_added
    node_dest = hexToFriendly(node.nodestack.verhex)
    wh, did = sdk_steward_wallet
    node_request = looper.loop.run_until_complete(
        prepare_node_request(did, node.name, destination=node_dest,
                             nodeIp=node.nodestack.ha[0],
                             nodePort=node.nodestack.ha[1],
                             clientIp=node.clientstack.ha[0],
                             clientPort=node.clientstack.ha[1]))
    return json.loads(node_request)


def ensurePoolIsOperable(looper, sdk_pool_handle, sdk_wallet_creator):
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_creator)


def testSendNodeFailsIfDestIsShortReadableName(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['dest'] = 'TheNewNode'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'b58 decoded value length 8 should be one of [16, 32]')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfDestIsHexKey(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['dest'] = cryptonymToHex(
        node_request['operation']['dest']).decode() + "0"
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'should not contain the following chars')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeHasInvalidSyntaxIfDestIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['dest'] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'b58 decoded value length 0 should be one of [16, 32]')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeHasInvalidSyntaxIfDestIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['dest']
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields - dest')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodeIpContainsLeadingSpace(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = ' 122.62.52.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodeIpContainsTrailingSpace(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = '122.62.52.13 '
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))

    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodeIpHasWrongFormat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = '122.62.52'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfSomeNodeIpComponentsAreNegative(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = '122.-1.52.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfSomeNodeIpComponentsAreHigherThanUpperBound(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = '122.62.256.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodeIpIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_IP] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodeIpIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][NODE_IP]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields - node_ip')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortIsNegative(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_PORT] = -1
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'network port out of the range 0-65535')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortIsHigherThanUpperBound(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_PORT] = 65536
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'network port out of the range 0-65535')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortIsFloat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_PORT] = 5555.5
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortHasWrongFormat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_PORT] = 'ninety'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][NODE_PORT] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types ')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfNodePortIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][NODE_PORT]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields - node_port')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientIpContainsLeadingSpace(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = ' 122.62.52.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientIpContainsTrailingSpace(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = '122.62.52.13 '
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientIpHasWrongFormat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = '122.62.52'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfSomeClientIpComponentsAreNegative(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = '122.-1.52.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfSomeClientIpComponentsAreHigherThanUpperBound(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = '122.62.256.13'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientIpIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_IP] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid network ip address')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientIpIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][CLIENT_IP]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields - client_ip')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortIsNegative(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_PORT] = -1
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'network port out of the range 0-65535')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortIsHigherThanUpperBound(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_PORT] = 65536
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'network port out of the range 0-65535')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortIsFloat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_PORT] = 5555.5
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortHasWrongFormat(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_PORT] = 'ninety'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][CLIENT_PORT] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfClientPortIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][CLIENT_PORT]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields - client_port')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfAliasIsEmpty(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][ALIAS] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'empty string')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfAliasIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][ALIAS]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields ')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfServicesContainsUnknownValue(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][SERVICES] = [VALIDATOR, 'DECIDER']
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'unknown value')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfServicesIsValidatorValue(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][SERVICES] = VALIDATOR  # just string, not array
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfServicesIsEmptyString(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][SERVICES] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'expected types')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeSuccessIfDataContainsUnknownField(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'][SERVICES] = []
    node_request['operation']['data']['extra'] = 42
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestRejectedException,
                         'action is not allowed')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfDataIsEmptyJson(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'] = {}
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields ')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfDataIsBrokenJson(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'] = "{'node_ip': '10.0.0.105', 'node_port': 9701"
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid type')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeFailsIfDataIsNotJson(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'] = 'not_json'
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid type')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeHasInvalidSyntaxIfDataIsEmptyString(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['data'] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid type')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeHasInvalidSyntaxIfDataIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data']
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'missed fields')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


@pytest.mark.skip(reason='INDY-1864')
def testSendNodeHasInvalidSyntaxIfUnknownParameterIsPassed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    node_request['operation']['albus'] = 'severus'
    steward_wallet, node = sdk_node_theta_added
    signed_reqs = sdk_sign_request_strings(looper, steward_wallet, [node_request])
    request_couple = sdk_send_signed_requests(sdk_pool_handle, signed_reqs)[0]
    sdk_get_and_check_replies(looper, [request_couple])
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeHasInvalidSyntaxIfAllParametersAreMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    for f in node_request['operation'].keys():
        node_request['operation'][f] = ''
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'invalid type')
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)


def testSendNodeSucceedsIfServicesIsMissed(
        looper, sdk_pool_handle, nodeSet, sdk_node_theta_added, node_request):
    del node_request['operation']['data'][SERVICES]
    steward_wallet, node = sdk_node_theta_added
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        json.dumps(node_request))
    sdk_get_and_check_replies(looper, [request_couple])
    ensurePoolIsOperable(looper, sdk_pool_handle, steward_wallet)



import pytest

from common.version import InvalidVersionError

from indy_common.constants import APP_NAME
from indy_common.version import (
    SchemaVersion, TopPkgDefVersion, NodeVersion, src_version_cls
)


def test_schema_version():
    for version in ['', '1', '1.2.3.4', '1.2.a']:
        with pytest.raises(InvalidVersionError):
            SchemaVersion(version)
    SchemaVersion('1.2')
    SchemaVersion('1.2.3')


def test_top_package_default_version():
    for version in ['', '1', '1.2.3.4', '1.2.a']:
        with pytest.raises(InvalidVersionError):
            TopPkgDefVersion(version)
    TopPkgDefVersion('1.2')
    TopPkgDefVersion('1.2.3')


def test_src_version_cls():
    assert src_version_cls() == NodeVersion
    assert src_version_cls(APP_NAME) == NodeVersion
    assert src_version_cls('some_package') == TopPkgDefVersion



import copy
import json
import random

import pytest

from indy_common.authorize.auth_constraints import AuthConstraintForbidden
from indy_common.constants import RS_ID, RS_TYPE, RS_NAME, RS_VERSION, RS_CONTENT, ENDORSER, JSON_LD_ID_FIELD, \
    JSON_LD_TYPE_FIELD
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_handler import RichSchemaHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_mapping_handler import \
    RichSchemaMappingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_pres_def_handler import \
    RichSchemaPresDefHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import context_request, make_rich_schema_object_exist
from plenum.common.constants import OP_VER, TRUSTEE
from plenum.common.exceptions import UnauthorizedClientRequest, InvalidClientRequest
from plenum.common.txn_util import reqToTxn, append_txn_metadata
from plenum.common.util import SortedDict, randomString


def test_update_state(handler_and_request):
    handler, request = handler_and_request
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(request)
    append_txn_metadata(txn, seq_no, txn_time)
    op = request.operation

    handler.update_state(txn, None, context_request)

    value = {
        'id': op[RS_ID],
        'rsType': op[RS_TYPE],
        'rsName': op[RS_NAME],
        'rsVersion': op[RS_VERSION],
        'content': op[RS_CONTENT],
        'from': request.identifier,
        'endorser': request.endorser,
        'ver': op[OP_VER],
    }
    primary_key = op[RS_ID]
    secondary_key = "{RS_TYPE}:{RS_NAME}:{RS_VERSION}".format(RS_TYPE=op['rsType'],
                                                              RS_NAME=op['rsName'],
                                                              RS_VERSION=op['rsVersion']).encode()

    value_from_state = handler.get_from_state(primary_key)
    assert SortedDict(value_from_state[0]) == SortedDict(value)
    assert value_from_state[1] == seq_no
    assert value_from_state[2] == txn_time
    assert handler.state.get(secondary_key, isCommitted=False) == op[RS_ID].encode()


def test_static_validation_pass(handler_and_request):
    handler, request = handler_and_request
    handler.static_validation(request)


def test_static_validation_content_is_json(handler_and_request):
    handler, request = handler_and_request

    request.operation[RS_CONTENT] = randomString()
    with pytest.raises(InvalidClientRequest, match="must be a JSON serialized string"):
        handler.static_validation(request)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_content_is_json_ld_with_atid(handler_and_request, status):
    handler, request = handler_and_request

    content = copy.deepcopy(json.loads(request.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop(JSON_LD_ID_FIELD, None)
    elif status == 'empty':
        content[JSON_LD_ID_FIELD] = ""
    elif status == 'none':
        content[JSON_LD_ID_FIELD] = None
    request.operation[RS_CONTENT] = json.dumps(content)

    if not isinstance(handler, (RichSchemaMappingHandler, RichSchemaHandler, RichSchemaPresDefHandler)):
        handler.static_validation(request)
        return

    with pytest.raises(InvalidClientRequest, match="'content' must be a valid JSON-LD and have non-empty '@id' field"):
        handler.static_validation(request)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_content_is_json_ld_with_attype(handler_and_request, status):
    handler, request = handler_and_request

    content = copy.deepcopy(json.loads(request.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop(JSON_LD_TYPE_FIELD, None)
    elif status == 'empty':
        content[JSON_LD_TYPE_FIELD] = ""
    elif status == 'none':
        content[JSON_LD_TYPE_FIELD] = None
    request.operation[RS_CONTENT] = json.dumps(content)

    if not isinstance(handler, (RichSchemaMappingHandler, RichSchemaHandler, RichSchemaPresDefHandler)):
        handler.static_validation(request)
        return

    with pytest.raises(InvalidClientRequest,
                       match="'content' must be a valid JSON-LD and have non-empty '@type' field"):
        handler.static_validation(request)


def test_static_validation_atid_equals_to_id(handler_and_request):
    handler, request = handler_and_request

    content = copy.deepcopy(json.loads(request.operation[RS_CONTENT]))
    content["@id"] = request.operation[RS_ID] + "a"
    request.operation[RS_CONTENT] = json.dumps(content)
    if not isinstance(handler, (RichSchemaMappingHandler, RichSchemaHandler, RichSchemaPresDefHandler)):
        handler.static_validation(request)
        return

    with pytest.raises(InvalidClientRequest,
                       match="content's @id must be equal to id={}".format(request.operation[RS_ID])):
        handler.static_validation(request)


def test_dynamic_validation_failed_not_authorised(handler_and_request):
    handler, request = handler_and_request
    add_to_idr(handler.database_manager.idr_cache, request.identifier, None)
    with pytest.raises(UnauthorizedClientRequest):
        handler.dynamic_validation(request, 0)


def test_dynamic_validation_for_existing(handler_and_request):
    handler, request = handler_and_request
    make_rich_schema_object_exist(handler, request)
    add_to_idr(handler.database_manager.idr_cache, request.identifier, TRUSTEE)
    add_to_idr(handler.database_manager.idr_cache, request.endorser, ENDORSER)

    with pytest.raises(UnauthorizedClientRequest, match=str(AuthConstraintForbidden())):
        handler.dynamic_validation(request, 0)


def test_dynamic_validation_for_existing_metadata(handler_and_request):
    handler, request = handler_and_request
    make_rich_schema_object_exist(handler, request)
    add_to_idr(handler.database_manager.idr_cache, request.identifier, TRUSTEE)
    add_to_idr(handler.database_manager.idr_cache, request.endorser, ENDORSER)

    request.operation[RS_ID] = randomString()
    request.operation[RS_CONTENT] = randomString()
    request.reqId = random.randint(10, 1000000000)

    with pytest.raises(InvalidClientRequest,
                       match='An object with rsName="{}", rsVersion="{}" and rsType="{}" already exists. '
                             'Please choose different rsName, rsVersion or rsType'.format(
                           request.operation[RS_NAME], request.operation[RS_VERSION], request.operation[RS_TYPE])):
        handler.dynamic_validation(request, 0)



import pytest

import indy_node
from indy_node.server.upgrade_log import UpgradeLog
from indy_common.constants import COMPLETE
from indy_node.test.upgrade.helper import populate_log_with_upgrade_events, \
    check_node_sent_acknowledges_upgrade, check_node_do_not_sent_acknowledges_upgrade, \
    emulate_restart_pool_for_upgrade, emulate_view_change_pool_for_upgrade

from indy_node.test.upgrade.helper import releaseVersion

whitelist = ['unable to send message']
# TODO: Implement a client in node

version = releaseVersion()


@pytest.fixture(scope="module")
def tdirWithPoolTxns(tdirWithPoolTxns, poolTxnNodeNames, tdir, tconf):
    # For each node, adding a file with he current version number which makes the node
    # think that an upgrade has been performed
    populate_log_with_upgrade_events(
        poolTxnNodeNames,
        tdir,
        tconf,
        version)
    return tdirWithPoolTxns


def test_node_detected_upgrade_done(nodeSet):
    '''
    Test that each node checks Upgrade Log on startup (after Upgrade restart), and writes SUCCESS to it
    because the current version equals to the one in Upgrade log.
    Upgrade log already created START event (see fixture above emulating real upgrade)
    '''
    for node in nodeSet:
        assert node.upgrader.scheduledAction is None
        assert node.upgrader.lastActionEventInfo.ev_type == UpgradeLog.Events.succeeded


def test_node_sent_upgrade_successful(looper, nodeSet, nodeIds):
    '''
    Test that each node sends NODE_UPGRADE Success event (because it sees SUCCESS in Upgrade log)
    Upgrade log already created START event (see fixture above emulating real upgrade)
    '''
    check_node_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                         allowed_actions=[COMPLETE],
                                         ledger_size=len(nodeSet),
                                         expected_version=version)


def test_node_sent_upgrade_successful_once_view_change(looper, nodeSet, nodeIds):
    '''
    Test that each node sends NODE_UPGRADE Success event only once after each view change
    '''
    # emulate view changes 1-4
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)

    # check that config ledger didn't changed (no new txns were sent)
    check_node_do_not_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                                allowed_actions=[COMPLETE],
                                                ledger_size=len(nodeSet),
                                                expected_version=version)


def test_node_sent_upgrade_successful_once_restart(looper, nodeSet, nodeIds):
    '''
    Test that each node sends NODE_UPGRADE Success event only once after restart,
    so that if we restart the node it's not sent again
    '''
    # emulate restart
    emulate_restart_pool_for_upgrade(nodeSet)

    # check that config ledger didn't changed (no new txns were sent)
    check_node_do_not_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                                allowed_actions=[COMPLETE],
                                                ledger_size=len(nodeSet),
                                                expected_version=version)



from copy import deepcopy

from indy_node.test import waits
from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION
from indy_common.constants import FORCE

from indy_node.test.upgrade.helper import bumpedVersion, checkUpgradeScheduled, \
    check_no_loop, sdk_ensure_upgrade_sent, clear_aq_stash
from indy_node.server.upgrade_log import UpgradeLog
from indy_node.utils.node_control_utils import NodeControlUtil

whitelist = ['Failed to upgrade node']


def test_upgrade_does_not_get_into_loop_force(looper, tconf, nodeSet,
                                              validUpgrade, sdk_pool_handle,
                                              sdk_wallet_trustee, monkeypatch):
    new_version = bumpedVersion(validUpgrade['version'])
    upgr1 = deepcopy(validUpgrade)
    upgr1[VERSION] = new_version
    upgr1[FORCE] = True

    clear_aq_stash(nodeSet)

    # An upgrade scheduled, it should pass
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))

    # here we make nodes think they have upgraded successfully
    monkeypatch.setattr(NodeControlUtil, '_get_curr_info',
                        lambda *x: "Version: {}".format(new_version))
    check_no_loop(nodeSet, UpgradeLog.Events.succeeded)



from indy_node.server.action_log import ActionLogData, ActionLogEvents, ActionLog


class RestartLogData(ActionLogData):
    pass


class RestartLog(ActionLog):

    Events = ActionLogEvents

    """
    Append-only event log of restart event
    """
    def __init__(self, file_path):
        super().__init__(
            file_path,
            data_class=RestartLogData,
            event_types=ActionLogEvents,
            delimiter='\t'
        )

    def append_scheduled(self, data: RestartLogData):
        super().append_scheduled(data)

    def append_started(self, data: RestartLogData):
        super().append_started(data)

    def append_succeeded(self, data: RestartLogData):
        super().append_succeeded(data)

    def append_failed(self, data: RestartLogData):
        super().append_failed(data)

    def append_cancelled(self, data: RestartLogData):
        super().append_cancelled(data)



import pytest

from indy_node.test.state_proof.helper import sdk_submit_operation_and_get_result
from plenum.common.constants import TARGET_NYM, RAW, NAME, VERSION, ORIGIN

# noinspection PyUnresolvedReferences
from plenum.common.ledger import Ledger
from plenum.test.primary_selection.test_view_changer_primary_selection import FakeLedger
from plenum.test.validator_info.conftest import info, node  # qa

from indy_common.constants import TXN_TYPE, DATA, GET_NYM, GET_ATTR, GET_SCHEMA, GET_CLAIM_DEF, REF, SIGNATURE_TYPE
from indy_node.__metadata__ import __version__ as node_pgk_version

PERIOD_SEC = 1
TEST_NODE_NAME = 'Alpha'
STATUS_FILENAME = '{}_info.json'.format(TEST_NODE_NAME.lower())
INFO_FILENAME = '{}_info.json'.format(TEST_NODE_NAME.lower())


def test_validator_info_file_schema_is_valid(info):
    assert isinstance(info, dict)
    assert 'config' in info['Node_info']['Metrics']['transaction-count']


def test_validator_info_file_metrics_count_ledger_field_valid(info):
    assert info['Node_info']['Metrics']['transaction-count']['config'] == 0


def test_validator_info_bls_key_field_valid(node, info):
    assert info['Node_info']['BLS_key']


def test_validator_info_ha_fields_valid(node, info):
    assert info['Node_info']['Node_ip']
    assert info['Node_info']['Client_ip']
    assert info['Node_info']['Node_port']
    assert info['Node_info']['Client_port']


def test_validator_info_file_software_indy_node_valid(info):
    assert info['Software']['indy-node'] == node_pgk_version


@pytest.fixture()
def node_with_broken_info_tool(node):
    node_bk = node._info_tool._node
    node._info_tool._node = None
    yield
    node._info_tool._node = node_bk


def test_validator_info_file_handle_fails(node_with_broken_info_tool,
                                          node):
    latest_info = node._info_tool.info

    assert 'Node_info' not in latest_info


def test_validator_info_file_get_nym(read_txn_and_get_latest_info,
                                     node):
    reset_node_total_read_request_number(node)
    info = node._info_tool.info
    assert info['Node_info']['Metrics']['average-per-second']['read-transactions'] == 0
    latest_info = read_txn_and_get_latest_info(GET_NYM)
    assert latest_info['Node_info']['Metrics']['average-per-second']['read-transactions'] > 0


def test_validator_info_file_get_schema(read_txn_and_get_latest_info,
                                        node):
    reset_node_total_read_request_number(node)
    info = node._info_tool.info
    assert info['Node_info']['Metrics']['average-per-second']['read-transactions'] == 0
    latest_info = read_txn_and_get_latest_info(GET_SCHEMA)
    assert latest_info['Node_info']['Metrics']['average-per-second']['read-transactions'] > 0


def test_validator_info_file_get_attr(read_txn_and_get_latest_info,
                                      node):
    reset_node_total_read_request_number(node)
    info = node._info_tool.info
    assert info['Node_info']['Metrics']['average-per-second']['read-transactions'] == 0
    latest_info = read_txn_and_get_latest_info(GET_ATTR)
    assert latest_info['Node_info']['Metrics']['average-per-second']['read-transactions'] > 0


def test_validator_info_file_get_claim_def(read_txn_and_get_latest_info,
                                           node):
    reset_node_total_read_request_number(node)
    info = node._info_tool.info
    assert info['Node_info']['Metrics']['average-per-second']['read-transactions'] == 0
    latest_info = read_txn_and_get_latest_info(GET_CLAIM_DEF)
    assert latest_info['Node_info']['Metrics']['average-per-second']['read-transactions'] > 0


def makeGetNymRequest(looper, sdk_pool_handle, sdk_wallet, nym):
    op = {
        TARGET_NYM: nym,
        TXN_TYPE: GET_NYM,
    }
    return sdk_submit_operation_and_get_result(looper, sdk_pool_handle, sdk_wallet, op)


def makeGetSchemaRequest(looper, sdk_pool_handle, sdk_wallet, nym):
    op = {
        TARGET_NYM: nym,
        TXN_TYPE: GET_SCHEMA,
        DATA: {
            NAME: 'foo',
            VERSION: '1.0',
        }
    }
    return sdk_submit_operation_and_get_result(looper, sdk_pool_handle, sdk_wallet, op)


def makeGetAttrRequest(looper, sdk_pool_handle, sdk_wallet, nym, raw):
    op = {
        TARGET_NYM: nym,
        TXN_TYPE: GET_ATTR,
        RAW: raw
    }
    return sdk_submit_operation_and_get_result(looper, sdk_pool_handle, sdk_wallet, op)


def makeGetClaimDefRequest(looper, sdk_pool_handle, sdk_wallet):
    op = {
        TXN_TYPE: GET_CLAIM_DEF,
        ORIGIN: '1' * 16,  # must be a valid DID
        REF: 1,
        SIGNATURE_TYPE: 'any'
    }
    return sdk_submit_operation_and_get_result(looper, sdk_pool_handle, sdk_wallet, op)


@pytest.fixture
def read_txn_and_get_latest_info(looper, sdk_pool_handle,
                                 sdk_wallet_client, node):
    _, did = sdk_wallet_client

    def read_wrapped(txn_type):
        if txn_type == GET_NYM:
            makeGetNymRequest(looper, sdk_pool_handle, sdk_wallet_client, did)
        elif txn_type == GET_SCHEMA:
            makeGetSchemaRequest(looper, sdk_pool_handle, sdk_wallet_client, did)
        elif txn_type == GET_ATTR:
            makeGetAttrRequest(looper, sdk_pool_handle, sdk_wallet_client, did, "attrName")
        elif txn_type == GET_CLAIM_DEF:
            makeGetClaimDefRequest(looper, sdk_pool_handle, sdk_wallet_client)
        else:
            assert False, "unexpected txn type {}".format(txn_type)
        return node._info_tool.info

    return read_wrapped


def reset_node_total_read_request_number(node):
    node.total_read_request_number = 0


class FakeTree:
    @property
    def root_hash(self):
        return '222222222222222222222222222'

class FakeLedgerEx(FakeLedger):
    @property
    def uncommittedRootHash(self):
        return '111111111111111111111111111111111'

    @property
    def uncommittedTxns(self):
        return []

    @property
    def tree(self):
        return FakeTree()

    @property
    def size(self):
        return 100


def test_validator_info_file_metrics_count_all_ledgers_field_valid(node):
    new_ids = [444, 555, 666, 777]
    for newid in new_ids:
        node.ledgerManager.addLedger(newid, FakeLedgerEx(newid, newid))
    info = node._info_tool.info
    has_cnt = len(info['Node_info']['Metrics']['transaction-count'])
    assert has_cnt == len(new_ids) + 4


def test_validator_info_update_date_field_valid(info):
    assert "Update_time" in info
    import time
    import datetime
    from_str = time.mktime(datetime.datetime.strptime(info["Update_time"],
                                                      "%A, %B %d, %Y %I:%M:%S %p %z").timetuple())
    assert int(from_str) == info["timestamp"]



import pytest
from indy_common.auth import Authoriser
from indy_common.authorize.auth_actions import EDIT_PREFIX, ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraintOr
from indy_common.authorize.auth_map import auth_map

from indy_common.constants import NYM, AUTH_RULE, OLD_VALUE, AUTH_TYPE, ROLE, ENDORSER, CONSTRAINT, AUTH_ACTION, \
    GET_AUTH_RULE

from indy_node.server.request_handlers.read_req_handlers.get_auth_rule_handler import GetAuthRuleHandler
from indy_node.test.auth_rule.helper import generate_auth_rule_operation, generate_key
from indy_node.test.request_handlers.helper import add_to_idr, get_exception
from plenum.common.constants import STEWARD, TRUSTEE, TXN_TYPE
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.request import Request
from plenum.common.util import randomString
from plenum.test.testing_utils import FakeSomething


@pytest.fixture(scope="module")
def get_auth_rule_handler(db_manager):
    f = FakeSomething()
    f.validate = lambda request, action_list: True
    f.auth_map = auth_map
    return GetAuthRuleHandler(db_manager, f)


@pytest.fixture(scope="function")
def get_auth_rule_request(creator):
    return Request(identifier=creator,
                   reqId=5,
                   operation={TXN_TYPE: GET_AUTH_RULE,
                              **generate_key()})


@pytest.fixture(scope="function")
def get_all_auth_rules_request(creator):
    return Request(identifier=creator,
                   reqId=5,
                   operation={TXN_TYPE: GET_AUTH_RULE})


@pytest.fixture(scope="module")
def creator(db_manager):
    identifier = randomString()
    idr = db_manager.idr_cache
    add_to_idr(idr, identifier, None)
    return identifier


def test_auth_rule_static_validation(get_auth_rule_request, get_auth_rule_handler: GetAuthRuleHandler):
    get_auth_rule_handler.static_validation(get_auth_rule_request)


def test_all_auth_rules_static_validation(get_all_auth_rules_request, get_auth_rule_handler: GetAuthRuleHandler):
    get_auth_rule_handler.static_validation(get_all_auth_rules_request)


def test_auth_rule_static_validation_failed_without_old_value(get_auth_rule_request,
                                                              get_auth_rule_handler: GetAuthRuleHandler):
    if OLD_VALUE in get_auth_rule_request.operation:
        del get_auth_rule_request.operation[OLD_VALUE]
    get_auth_rule_request.operation[AUTH_ACTION] = EDIT_PREFIX
    with pytest.raises(InvalidClientRequest, match="EDIT must contain field old_value"):
        get_auth_rule_handler.static_validation(get_auth_rule_request)


def test_auth_rule_static_validation_failed_with_excess_field(get_auth_rule_request,
                                                              get_auth_rule_handler: GetAuthRuleHandler):
    get_auth_rule_request.operation[OLD_VALUE] = "old_value"
    get_auth_rule_request.operation[AUTH_ACTION] = ADD_PREFIX
    with pytest.raises(InvalidClientRequest, match="ADD must not contain field old_value"):
        get_auth_rule_handler.static_validation(get_auth_rule_request)


def test_auth_rule_static_validation_failed_with_incorrect_key(get_auth_rule_request,
                                                               get_auth_rule_handler: GetAuthRuleHandler):
    get_auth_rule_request.operation.update(generate_key(auth_action=EDIT_PREFIX,
                                                        auth_type="wrong_type",
                                                        field=ROLE,
                                                        new_value=ENDORSER,
                                                        old_value="*"))
    with pytest.raises(InvalidClientRequest, match="is not found in authorization map"):
        get_auth_rule_handler.static_validation(get_auth_rule_request)



import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX, AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.constants import ROLE
from indy_node.test.auth_rule.test_auth_rule_transaction import sdk_send_and_check_auth_rule_request
from indy_node.test.auth_rule.test_check_rule_for_add_action_changing import create_verkey_did
from plenum.common.constants import STEWARD, NYM, STEWARD_STRING, PREPARE, COMMIT, DOMAIN_LEDGER_ID, CONFIG_LEDGER_ID
from plenum.common.exceptions import RequestRejectedException
from plenum.common.startable import Mode
from plenum.test import waits
from plenum.test.delayers import cDelay, pDelay, msg_rep_delay
from plenum.test.helper import assertExp
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.stasher import delay_rules_without_processing
from plenum.test.view_change.helper import ensure_view_change_complete
from stp_core.loop.eventually import eventually


def test_revert_auth_rule_changing(looper,
                                   txnPoolNodeSet,
                                   sdk_wallet_trustee,
                                   sdk_wallet_steward,
                                   sdk_pool_handle):
    node_stashers = [n.nodeIbStasher for n in txnPoolNodeSet]
    wh, _ = sdk_wallet_trustee
    new_steward_did, new_steward_verkey = create_verkey_did(looper, wh)
    new_steward_did2, new_steward_verkey2 = create_verkey_did(looper, wh)
    """We try to change rule for adding new steward. For this case we """
    changed_constraint = AuthConstraint(role=STEWARD,
                                        sig_count=1)
    action = AuthActionAdd(txn_type=NYM,
                           field=ROLE,
                           value=STEWARD)
    with delay_rules_without_processing(node_stashers, pDelay(), cDelay(),
                                        msg_rep_delay(types_to_delay=[PREPARE, COMMIT])):
        sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                             auth_action=ADD_PREFIX,
                                             auth_type=action.txn_type,
                                             field=action.field,
                                             new_value=action.value,
                                             old_value=None,
                                             constraint=changed_constraint.as_dict,
                                             no_wait=True)
        looper.runFor(waits.expectedPrePrepareTime(len(txnPoolNodeSet)))
        """
        Try to add new steward by already existed trustee.
        Validation should raise exception because we change uncommitted state
        by adding new rule, that "Only steward can add new steward"
        """
        with pytest.raises(RequestRejectedException, match="Not enough STEWARD signatures"):
            sdk_add_new_nym(looper,
                            sdk_pool_handle,
                            sdk_wallet_trustee,
                            'newSteward1',
                            STEWARD_STRING,
                            dest=new_steward_did, verkey=new_steward_verkey)
        looper.runFor(waits.expectedPrePrepareTime(len(txnPoolNodeSet)))
        """
        Catchup should revert config_state and discard rule changing
        """
        for n in txnPoolNodeSet:
            n.start_catchup()
        # clear all request queues to not re-send the AuthRule txns again
        for n in txnPoolNodeSet:
            n.requests.clear()
            for r in n.replicas.values():
                r._ordering_service.requestQueues[DOMAIN_LEDGER_ID].clear()
                r._ordering_service.requestQueues[CONFIG_LEDGER_ID].clear()
        looper.run(
            eventually(lambda nodes: assertExp(all([n.mode == Mode.participating for n in nodes])), txnPoolNodeSet))

    # do view change to not send PrePrepares with the same ppSeqNo and viewNo
    ensure_view_change_complete(looper, txnPoolNodeSet)

    """
    Try to create new steward by steward
    We can not do this, because AUTH_RULE txn was reverted
    """
    with pytest.raises(RequestRejectedException, match="Not enough TRUSTEE signatures"):
        sdk_add_new_nym(looper,
                        sdk_pool_handle,
                        sdk_wallet_steward,
                        'newSteward2',
                        STEWARD_STRING,
                        dest=new_steward_did2, verkey=new_steward_verkey2)
    # """But with rules before sending AUTH_RULE txn (by default) we can add new steward"""
    # sdk_add_new_nym(looper,
    #                 sdk_pool_handle,
    #                 sdk_wallet_trustee,
    #                 'newSteward3',
    #                 STEWARD_STRING,
    #                 dest=new_steward_did2, verkey=new_steward_verkey2)



from indy_common.authorize.helper import get_named_role
from plenum.common.constants import TRUSTEE_STRING, TRUSTEE


def test_for_known_role():
    assert get_named_role(TRUSTEE) == TRUSTEE_STRING


def test_for_unknown_role():
    assert get_named_role("SomeOtherRole") == "Unknown role"



import pytest
import os
from datetime import datetime
from dateutil.parser import parse as parse_dt
from enum import Enum, unique

from indy_node.server.action_log import (
    CsvSerializer,
    ActionLogEvents, ActionLogData, ActionLogEvent, ActionLog,
)

# TODO
# - check specific error messages for expected exceptions
# - try to load current logs
# - move csv serializer tests to separate module


class LogDataParent(CsvSerializer):
    _items = ['f1']

    def __init__(self, f1):
        self.f1 = f1


class LogData(LogDataParent):
    _items = LogDataParent._items + ['f2']

    def __init__(self, f1, f2):
        super().__init__(f1)
        self.f2 = int(f2)


@unique
class Events(Enum):
    eventA = 1
    eventB = 2
    eventC = 3

    def __str__(self):
        return self.name


class ActionTestLog(ActionLog):

    Events = Events

    """
    Append-only event log of upgrade event
    """
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            data_class=LogData,
            event_types=Events,
            **kwargs
        )


# FIXTURES


@pytest.fixture
def valid_ts():
    return datetime.utcnow()


@pytest.fixture(scope='module')
def valid_event():
    return Events.eventA


@pytest.fixture(scope='module')
def valid_data():
    return LogData('1', 2)


@pytest.fixture(scope='module')
def log_delimiter(tdir, request):
    return '>'


@pytest.fixture
def log_file_path(tdir, request):
    return os.path.join(
        tdir,
        "{}.action_log".format(os.path.basename(request.node.nodeid))
    )


@pytest.fixture(scope='module')
def prepared_data():
    return [
        LogData('1', 1),
        LogData('2', 2),
        LogData('3', 3)
    ]


@pytest.fixture
def action_log(log_file_path, log_delimiter):
    return ActionTestLog(log_file_path, delimiter=log_delimiter)


@pytest.fixture
def action_log_prepared(action_log, prepared_data):
    for data in prepared_data:
        action_log.append_eventB(data)
    return action_log


class CsvSerializerTest(CsvSerializer):
    _items = ['int_', 'ts', 'str_']

    def __init__(self, int_, ts, str_, *args, **kwargs):
        self.int_ = int(int_)
        self.ts = parse_dt(ts)
        self.str_ = str_
        self.args = args
        self.kwargs = kwargs

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other


@pytest.fixture(scope='module')
def test_obj():
    return CsvSerializerTest('1', str(datetime.utcnow()), 'qwerty')


# TESTS


def test_csv_serializer_iter(test_obj):
    assert list(iter(test_obj)) == [test_obj.int_, test_obj.ts, test_obj.str_]


def test_csv_serializer_pack_unpack():
    delimiter = '_'
    args = (5,)
    kwargs = {'kw': 6}
    test_obj = CsvSerializerTest(
        '1', str(datetime.utcnow()), 'qwerty', *args, **kwargs
    )
    assert test_obj == CsvSerializerTest.unpack(
        test_obj.pack(delimiter=delimiter), *args, delimiter=delimiter, **kwargs
    )


def test_action_log_data_init(valid_ts):
    assert ActionLogData(valid_ts).when == valid_ts
    assert ActionLogData(valid_ts.isoformat()).when == valid_ts
    assert ActionLogData(str(valid_ts)).when == valid_ts
    with pytest.raises(TypeError):
        ActionLogData(123)


def test_action_log_event_init_invalid_ts(valid_ts, valid_event):
    with pytest.raises(TypeError):
        ActionLogEvent(
            123, valid_event, ActionLogData(datetime.utcnow()),
        )


def test_action_log_event_init_valid_ts_none(valid_event, valid_data):
    ts1 = datetime.utcnow()
    ev = ActionLogEvent(None, valid_event, valid_data, types=Events)
    ts2 = datetime.utcnow()
    # a kind of non-strict check
    assert ts1 <= ev.ts <= ts2


def test_action_log_event_init_valid_ts_datetime(
        valid_ts, valid_event, valid_data):
    ev = ActionLogEvent(valid_ts, valid_event, valid_data, types=Events)
    assert ev.ts == valid_ts


def test_action_log_event_init_valid_ts_str(
        valid_ts, valid_event, valid_data):
    ev = ActionLogEvent(valid_ts.isoformat(), valid_event,
                        valid_data, types=Events)
    assert ev.ts == valid_ts


def test_action_log_event_init_invalid_ev_type(valid_ts, valid_data):
    _Events = Enum('_Events', 'ev1 ev2 ev3')

    # unexpected type
    with pytest.raises(TypeError):
        ActionLogEvent(valid_ts, 123, valid_data, types=_Events)

    # unexpected event name
    with pytest.raises(ValueError):
        ActionLogEvent(valid_ts, 'ev4', valid_data, types=_Events)

    # unexpected event
    with pytest.raises(ValueError):
        ActionLogEvent(
            valid_ts, Enum('_Events', 'ev1 ev2 ev3')['ev1'],
            valid_data, types=_Events
        )


def test_action_log_event_init_valid_ev_type(valid_ts, valid_data):
    _Events = Enum('_Events', 'ev1 ev2 ev3')

    assert ActionLogEvent(
        valid_ts, 'ev2', valid_data, types=_Events
    ).ev_type == _Events.ev2

    assert ActionLogEvent(
        valid_ts, _Events.ev3, valid_data, types=_Events
    ).ev_type == _Events.ev3


def test_action_log_event_init_invalid_data(valid_ts, valid_event):
    # data_class is not specified and data not an instance of CsvSerializer
    with pytest.raises(TypeError):
        ActionLogEvent(valid_ts, valid_event, valid_ts, types=Events)

    # data_class is specified but it's not a CsvSerializer child
    with pytest.raises(TypeError):
        ActionLogEvent(
            valid_ts, valid_event, 'any_value', types=Events,
            data_class=type("SomeClass", (), {'__init__': lambda *_: None})
        )


def test_action_log_event_init_valid_data(valid_ts, valid_event, valid_data):
    assert ActionLogEvent(
        valid_ts, valid_event, valid_data, types=Events
    ).data == valid_data


def test_action_log_event_init_data_class_passed(valid_ts, valid_event):

    class SomeClass(CsvSerializer):
        _items = ['item1', 'item2', 'item3']

        def __init__(self, *args):
            for idx, v in enumerate(args):
                setattr(self, "item" + str(idx), v)

    ev = ActionLogEvent(
        valid_ts, valid_event, '_item1', '_item2', '_item3',
        types=Events, data_class=SomeClass
    )

    ev.data == SomeClass('_item1', '_item2', '_item3')


def test_action_log_event_pack_unpack(valid_ts, valid_event, valid_data):
    delimiter = '|'
    ev = ActionLogEvent(valid_ts, valid_event, valid_data, types=Events)
    assert ev == ActionLogEvent.unpack(
        ev.pack(delimiter=delimiter), delimiter=delimiter,
        data_class=LogData, types=Events)


def test_action_log_api_basic(
        action_log_prepared, log_file_path,
        log_delimiter, prepared_data):

    # basic ones
    assert action_log_prepared.file_path == log_file_path
    assert action_log_prepared.delimiter == log_delimiter
    assert len(action_log_prepared) == len(prepared_data)
    for ev, data in zip(action_log_prepared, prepared_data):
        assert ev.data == data
    assert action_log_prepared.last_event.data == prepared_data[-1]

    # helpers for append
    for ev_type in Events:
        assert hasattr(action_log_prepared, "append_{}".format(ev_type.name))


@pytest.mark.parametrize('ev_type', Events)
def test_action_log_append(action_log, ev_type):
    ev_data = LogData('1', 2)
    getattr(action_log, "append_{}".format(ev_type.name))(ev_data)
    assert action_log.last_event.data == ev_data
    assert action_log.last_event.ev_type == ev_type


def test_action_log_write_file(action_log):
    ev_data = LogData('1', 2)

    def check_log(expected):
        num_lines = -1
        with open(action_log.file_path, 'r') as _f:
            for num_lines, _ in enumerate(_f):
                pass
        assert (num_lines + 1) == expected

    action_log.append_eventA(ev_data)
    check_log(1)
    action_log.append_eventB(ev_data)
    check_log(2)
    action_log.append_eventC(ev_data)
    check_log(3)


def test_action_log_load(action_log_prepared):
    new_log = ActionTestLog(
        action_log_prepared.file_path,
        delimiter=action_log_prepared.delimiter
    )

    assert len(action_log_prepared) == len(new_log)
    for ev1, ev2 in zip(action_log_prepared, new_log):
        assert ev1 == ev2
